"""
Model(s) for grouped_light_level resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light_level
"""

from dataclasses import dataclass

from aiohue.v2.models.light_level import LightLevelFeature

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class GroupedLightLevel:
    """
    Represent a (full) `GroupedLightLevel` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light_level_get
    """

    id: str
    owner: ResourceIdentifier
    enabled: bool
    light: LightLevelFeature

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.GROUPED_LIGHT_LEVEL


@dataclass
class GroupedLightLevelPut:
    """
    GroupedLightLevel resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light_level__id__put
    """

    enabled: bool | None = None
    type: ResourceTypes = ResourceTypes.GROUPED_LIGHT_LEVEL



"""Controller holding and managing HUE resources of type `light`."""

from aiohue.v2.models.feature import (
    AlertEffectType,
    AlertFeaturePut,
    ColorFeaturePut,
    ColorPoint,
    ColorTemperatureFeaturePut,
    DeltaAction,
    DimmingFeaturePut,
    DimmingDeltaFeaturePut,
    DynamicsFeaturePut,
    EffectsFeaturePut,
    EffectStatus,
    OnFeature,
    TimedEffectsFeaturePut,
    TimedEffectStatus,
)
from aiohue.v2.models.light import Light, LightPut
from aiohue.v2.models.resource import ResourceTypes

from .base import BaseResourcesController


class LightsController(BaseResourcesController[type[Light]]):
    """Controller holding and managing HUE resources of type `light`."""

    item_type = ResourceTypes.LIGHT
    item_cls = Light

    async def turn_on(self, id: str, transition_time: int | None = None) -> None:
        """Turn on the light."""
        await self.set_state(id, on=True, transition_time=transition_time)

    async def turn_off(self, id: str, transition_time: int | None = None) -> None:
        """Turn off the light."""
        await self.set_state(id, on=False, transition_time=transition_time)

    async def set_brightness(
        self, id: str, brightness: float, transition_time: int | None = None
    ) -> None:
        """Set brightness to light. Turn on light if it's currently off."""
        await self.set_state(
            id, on=True, brightness=brightness, transition_time=transition_time
        )

    async def set_color(
        self, id: str, x: float, y: float, transition_time: int | None = None
    ) -> None:
        """Set color to light. Turn on light if it's currently off."""
        await self.set_state(
            id, on=True, color_xy=(x, y), transition_time=transition_time
        )

    async def set_color_temperature(
        self, id: str, mirek: int, transition_time: int | None = None
    ) -> None:
        """Set Color Temperature to light. Turn on light if it's currently off."""
        await self.set_state(
            id, on=True, color_temp=mirek, transition_time=transition_time
        )

    async def set_flash(self, id: str, short: bool = False) -> None:
        """Send Flash command to light."""
        if short:
            device = self.get_device(id)
            await self._bridge.devices.set_identify(device.id)
        else:
            await self.set_state(id, alert=AlertEffectType.BREATHE)

    async def set_state(
        self,
        id: str,
        on: bool | None = None,
        brightness: float | None = None,
        color_xy: tuple[float, float] | None = None,
        color_temp: int | None = None,
        transition_time: int | None = None,
        alert: AlertEffectType | None = None,
        effect: EffectStatus | TimedEffectStatus | None = None,
    ) -> None:
        """Set supported feature(s) to light resource."""
        update_obj = LightPut()
        if on is not None:
            update_obj.on = OnFeature(on=on)
        if brightness is not None:
            update_obj.dimming = DimmingFeaturePut(brightness=brightness)
        if color_xy is not None:
            update_obj.color = ColorFeaturePut(xy=ColorPoint(*color_xy))
        if color_temp is not None:
            update_obj.color_temperature = ColorTemperatureFeaturePut(mirek=color_temp)
        if transition_time is not None and effect in (None, EffectStatus.NO_EFFECT):
            update_obj.dynamics = DynamicsFeaturePut(duration=transition_time)
        if alert is not None:
            update_obj.alert = AlertFeaturePut(action=alert)

        # for timed_effects feature, transition time is used for duration
        if effect is not None and isinstance(effect, TimedEffectStatus):
            update_obj.timed_effects = TimedEffectsFeaturePut(
                effect=effect, duration=transition_time
            )
        elif effect is not None:
            update_obj.effects = EffectsFeaturePut(effect=effect)
        await self.update(id, update_obj)

    async def set_dimming_delta(
        self, id: str, brightness_delta: float | None = None
    ) -> None:
        """
        Set brightness_delta value and action via DimmingDeltaFeature.

        The action to be send depends on brightness_delta value:
         None: STOP (this immediately stops any dimming transition)
         > 0: UP,
         < 0: DOWN
        """
        if brightness_delta in (None, 0):
            dimming_delta = DimmingDeltaFeaturePut(action=DeltaAction.STOP)
        else:
            dimming_delta = DimmingDeltaFeaturePut(
                action=DeltaAction.UP if brightness_delta > 0 else DeltaAction.DOWN,
                brightness_delta=abs(brightness_delta),
            )

        update_obj = LightPut()
        update_obj.dimming_delta = dimming_delta
        await self.update(id, update_obj)



"""Controller holding and managing HUE resources of type `device`."""

from aiohue.v2.models.device import Device, DevicePut
from aiohue.v2.models.feature import IdentifyFeature
from aiohue.v2.models.light import Light
from aiohue.v2.models.resource import ResourceTypes
from aiohue.v2.models.room import Room
from aiohue.v2.models.zigbee_connectivity import ZigbeeConnectivity

from .base import BaseResourcesController
from .sensors import SENSOR_TYPES


class DevicesController(BaseResourcesController[type[Device]]):
    """Controller holding and managing HUE resources of type `device`."""

    item_type = ResourceTypes.DEVICE
    item_cls = Device

    def get_lights(self, id: str) -> list[Light]:
        """Return all lights belonging to given device."""
        return [x for x in self._bridge.lights if x.id in self[id].lights]

    def get_sensors(self, id: str) -> list[SENSOR_TYPES]:
        """Return all sensors belonging to given device."""
        return [x for x in self._bridge.sensors if x.id in self[id].sensors]

    def get_room(self, id: str) -> Room | None:
        """Return room the given device belongs to (if any)."""
        return next((x for x in self._bridge.groups.room if id in x.devices), None)

    def get_zigbee_connectivity(self, id: str) -> ZigbeeConnectivity | None:
        """Return the ZigbeeConnectivity resource connected to device."""
        for service in self._items[id].services:
            if service.rtype == ResourceTypes.ZIGBEE_CONNECTIVITY:
                return self._bridge.sensors.zigbee_connectivity.get(service.rid)
        return None

    async def set_identify(self, id: str) -> None:
        """
        Send identify command to the device.

        Triggers a visual identification sequence, current implemented as
        (which can change in the future):
        - Bridge performs Zigbee LED identification cycles for 5 seconds
        - Lights perform one breathe cycle
        - Sensors perform LED identification cycles for 15 seconds
        """
        await self.update(id, DevicePut(identify=IdentifyFeature()))



"""
Model(s) for service_group resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_service_group
"""

from dataclasses import dataclass

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class ServiceGroupMetadata:
    """
    Represent metadata for service group.

    Used by `service_group` resources.
    """

    name: str


@dataclass
class ServiceGroup:
    """
    Represent a (full) `ServiceGroup` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_service_group_get
    """

    id: str
    children: list[ResourceIdentifier]
    services: list[ResourceIdentifier]
    metadata: ServiceGroupMetadata

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.SERVICE_GROUP


@dataclass
class ServiceGroupPut:
    """
    ServiceGroup resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_service_group__id__put
    """

    metadata: ServiceGroupMetadata | None = None



"""
Model(s) for zigbee_connectivity resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zgp_connectivity
"""

from dataclasses import dataclass

from .resource import ResourceIdentifier, ResourceTypes
from .zigbee_connectivity import ConnectivityServiceStatus


@dataclass
class ZgpConnectivity:
    """
    Represent a (full) `ZgpConnectivity` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zgp_connectivity_get
    """

    id: str
    owner: ResourceIdentifier
    status: ConnectivityServiceStatus
    source_id: str
    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.ZGP_CONNECTIVITY



"""Test button controller functions."""

import asyncio
from copy import deepcopy
from unittest.mock import Mock
from uuid import uuid4

from aiohue import HueBridgeV2
from aiohue.v2 import EventType
from aiohue.v2.controllers.sensors import ButtonController
from aiohue.v2.models.button import ButtonEvent
from aiohue.v2.models.resource import ResourceTypes


async def create_button(
    controller: ButtonController, callback: Mock, button_id: str, device_id: str
):
    """Create a button resource."""
    evt_data = {
        "id": button_id,
        "metadata": {"control_id": 1},
        "button": {
            "event_values": [
                "initial_press",
                "repeat",
                "short_release",
                "long_release",
                "long_press",
            ],
            "repeat_interval": 800,
        },
        "owner": {"rid": device_id, "rtype": ResourceTypes.DEVICE},
    }

    # Create a new resource
    # pylint: disable=protected-access
    await controller._handle_event(EventType.RESOURCE_ADDED, evt_data)

    callback.assert_called_once()
    callback.reset_mock()


async def create_device(
    bridge: HueBridgeV2, device_id: str, model_id: str, button_id: str
):
    """Create a device resource."""
    evt_data = {
        "id": device_id,
        "services": [{"rid": button_id, "rtype": "button"}],
        "product_data": {
            "model_id": model_id,
            "manufacturer_name": "Signify",
            "product_name": "Hue Light",
            "product_archetype": "classic_bulb",
            "certified": True,
            "software_version": "1.0.0",
        },
        "metadata": {"archetype": "classic_bulb", "name": "Kitchen"},
    }

    # pylint: disable=protected-access
    await bridge.devices._handle_event(EventType.RESOURCE_ADDED, evt_data)


def generate_button_event_data(button_id: str, event: str, device_id: str):
    """Generate button event data."""
    return {
        "id": button_id,
        "button": {
            "button_report": {
                "event": event,
                "updated": "2024-08-24T16:24:00.0Z",
            }
        },
        "owner": {"rid": device_id, "rtype": ResourceTypes.DEVICE},
    }


async def handle_button_event(
    controller: ButtonController, button_id: str, event: str, device_id: str
):
    """Handle button event."""
    # pylint: disable=protected-access
    await controller._handle_event(
        EventType.RESOURCE_UPDATED,
        generate_button_event_data(button_id, event, device_id),
    )


class CopyingMock(Mock):
    """Mock that deep copies its arguments."""

    def __call__(self, *args, **kwargs):
        args = deepcopy(args)
        kwargs = deepcopy(kwargs)
        return super().__call__(*args, **kwargs)


async def test_handle_events():
    """Test handling of events."""
    bridge = HueBridgeV2("127.0.0.1", "fake")
    controller = ButtonController(bridge)
    callback = CopyingMock(return_value=None)
    controller.subscribe(callback)

    button_id = str(uuid4())
    device_id = str(uuid4())

    await create_button(controller, callback, button_id, device_id)

    # Button event
    await handle_button_event(controller, button_id, "initial_press", device_id)

    callback.assert_called_once()
    assert (
        callback.call_args.args[1].button.button_report.event
        == ButtonEvent.INITIAL_PRESS
    )
    callback.reset_mock()

    evt_data = {
        "id": button_id,
        "owner": {"rid": device_id, "rtype": ResourceTypes.DEVICE},
    }

    # Absent button report is dropped
    # pylint: disable=protected-access
    await controller._handle_event(EventType.RESOURCE_UPDATED, evt_data)

    callback.assert_not_called()
    callback.reset_mock()


async def test_handle_events_button_workaround_short_release():
    """Test handling of events for device requiring button workaround."""
    bridge = HueBridgeV2("127.0.0.1", "fake")
    controller = ButtonController(bridge)
    callback = CopyingMock(return_value=None)
    controller.subscribe(callback)

    button_id = str(uuid4())
    device_id = str(uuid4())

    await create_device(bridge, device_id, "FOHSWITCH", button_id)
    await create_button(controller, callback, button_id, device_id)

    # Button event: initial_press
    await handle_button_event(controller, button_id, "initial_press", device_id)

    await asyncio.sleep(1.2)

    # Button event: short_release
    await handle_button_event(controller, button_id, "short_release", device_id)

    callback.assert_called()
    assert callback.call_count == 2
    assert (
        callback.call_args_list[0].args[1].button.button_report.event
        == ButtonEvent.INITIAL_PRESS
    )
    assert (
        callback.call_args_list[1].args[1].button.button_report.event
        == ButtonEvent.SHORT_RELEASE
    )


async def test_handle_events_button_workaround_repeats():
    """Test handling of events for device requiring button workaround."""
    bridge = HueBridgeV2("127.0.0.1", "fake")
    controller = ButtonController(bridge)
    callback = CopyingMock(return_value=None)
    controller.subscribe(callback)

    button_id = str(uuid4())
    device_id = str(uuid4())

    await create_device(bridge, device_id, "FOHSWITCH", button_id)
    await create_button(controller, callback, button_id, device_id)

    # Button event: initial_press
    await handle_button_event(controller, button_id, "initial_press", device_id)

    await asyncio.sleep(2.2)

    # Button event: short_release
    await handle_button_event(controller, button_id, "short_release", device_id)

    callback.assert_called()
    assert callback.call_count == 4
    assert (
        callback.call_args_list[0].args[1].button.button_report.event
        == ButtonEvent.INITIAL_PRESS
    )
    assert (
        callback.call_args_list[1].args[1].button.button_report.event
        == ButtonEvent.REPEAT
    )
    assert (
        callback.call_args_list[2].args[1].button.button_report.event
        == ButtonEvent.REPEAT
    )
    assert (
        callback.call_args_list[3].args[1].button.button_report.event
        == ButtonEvent.SHORT_RELEASE
    )


async def test_handle_events_button_workaround_long_release():
    """Test handling of events for device requiring button workaround."""
    bridge = HueBridgeV2("127.0.0.1", "fake")
    controller = ButtonController(bridge)
    callback = CopyingMock(return_value=None)
    controller.subscribe(callback)

    button_id = str(uuid4())
    device_id = str(uuid4())

    await create_device(bridge, device_id, "FOHSWITCH", button_id)
    await create_button(controller, callback, button_id, device_id)

    # Button event: initial_press
    await handle_button_event(controller, button_id, "initial_press", device_id)

    await asyncio.sleep(12.7)

    callback.assert_called()
    assert callback.call_count == 23
    assert (
        callback.call_args_list[22].args[1].button.button_report.event
        == ButtonEvent.LONG_RELEASE
    )


async def test_handle_events_button_workaround_interrupt():
    """Test handling of events for device requiring button workaround."""
    bridge = HueBridgeV2("127.0.0.1", "fake")
    controller = ButtonController(bridge)
    callback = CopyingMock(return_value=None)
    controller.subscribe(callback)

    button1_id = str(uuid4())
    button2_id = str(uuid4())
    device1_id = str(uuid4())
    device2_id = str(uuid4())

    await create_device(bridge, device1_id, "FOHSWITCH", button1_id)
    await create_device(bridge, device2_id, "ZGPSWITCH", button2_id)
    await create_button(controller, callback, button1_id, device1_id)
    await create_button(controller, callback, button2_id, device2_id)

    # Button event: initial_press
    await handle_button_event(controller, button1_id, "initial_press", device1_id)

    await asyncio.sleep(2.2)

    # Button event: initial_press
    await handle_button_event(controller, button2_id, "initial_press", device2_id)

    callback.assert_called()
    assert callback.call_count == 4
    assert (
        callback.call_args_list[0].args[1].button.button_report.event
        == ButtonEvent.INITIAL_PRESS
    )
    assert (
        callback.call_args_list[1].args[1].button.button_report.event
        == ButtonEvent.REPEAT
    )
    assert (
        callback.call_args_list[2].args[1].button.button_report.event
        == ButtonEvent.REPEAT
    )
    assert callback.call_args_list[3].args[1].id == button2_id
    assert (
        callback.call_args_list[3].args[1].button.button_report.event
        == ButtonEvent.INITIAL_PRESS
    )


async def test_handle_events_button_non_workaround_device():
    """Test handling of events for device requiring button workaround."""
    bridge = HueBridgeV2("127.0.0.1", "fake")
    controller = ButtonController(bridge)
    callback = CopyingMock(return_value=None)
    controller.subscribe(callback)

    button_id = str(uuid4())
    device_id = str(uuid4())

    await create_device(bridge, device_id, "ZGPSWITCH", button_id)
    await create_button(controller, callback, button_id, device_id)

    # Button event: initial_press
    await handle_button_event(controller, button_id, "initial_press", device_id)

    await asyncio.sleep(2.2)

    # Button event: short_release
    await handle_button_event(controller, button_id, "short_release", device_id)

    callback.assert_called()
    assert callback.call_count == 2
    assert (
        callback.call_args_list[0].args[1].button.button_report.event
        == ButtonEvent.INITIAL_PRESS
    )
    assert (
        callback.call_args_list[1].args[1].button.button_report.event
        == ButtonEvent.SHORT_RELEASE
    )



"""Test helpers for v2."""

import pytest

from tests.common import load_fixture


@pytest.fixture(scope="session")
def v2_resources():
    """Load v2 resources."""
    return load_fixture("v2_resources.json")



"""
Model(s) for security_area_motion resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_security_area_motion
"""

from dataclasses import dataclass

from .feature import (
    MotionSensingFeature,
    MotionSensingFeatureSensitivity,
    MotionSensingFeatureSensitivityPut,
)
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class SecurityAreaMotion:
    """
    Represent a (full) `SecurityAreaMotion` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_security_area_motion_get
    """

    id: str
    owner: ResourceIdentifier
    # enabled: required(boolean)
    # true when sensor is activated, false when deactivated
    enabled: bool
    motion: MotionSensingFeature
    sensitivity: MotionSensingFeatureSensitivity | None

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.SECURITY_AREA_MOTION


@dataclass
class SecurityAreaMotionPut:
    """
    SecurityAreaMotion resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_security_area_motion__id__put
    """

    enabled: bool | None = None
    sensitivity: MotionSensingFeatureSensitivityPut | None = None



"""Hue sensor resources."""

from __future__ import annotations

from .api import APIItems

TYPE_DAYLIGHT = "Daylight"

TYPE_CLIP_GENERICFLAG = "CLIPGenericFlag"
TYPE_CLIP_GENERICSTATUS = "CLIPGenericStatus"
TYPE_CLIP_HUMIDITY = "CLIPHumidity"
TYPE_CLIP_LIGHTLEVEL = "CLIPLightLevel"
TYPE_CLIP_OPENCLOSE = "CLIPOpenClose"
TYPE_CLIP_PRESENCE = "CLIPPresence"
TYPE_CLIP_SWITCH = "CLIPSwitch"
TYPE_CLIP_TEMPERATURE = "CLIPTemperature"

TYPE_GEOFENCE = "Geofence"

TYPE_ZGP_SWITCH = "ZGPSwitch"

TYPE_ZLL_LIGHTLEVEL = "ZLLLightLevel"
TYPE_ZLL_PRESENCE = "ZLLPresence"
TYPE_ZLL_ROTARY = "ZLLRelativeRotary"
TYPE_ZLL_SWITCH = "ZLLSwitch"
TYPE_ZLL_TEMPERATURE = "ZLLTemperature"

ZGP_SWITCH_BUTTON_1 = 34
ZGP_SWITCH_BUTTON_2 = 16
ZGP_SWITCH_BUTTON_3 = 17
ZGP_SWITCH_BUTTON_4 = 18

ZLL_SWITCH_BUTTON_1_INITIAL_PRESS = 1000
ZLL_SWITCH_BUTTON_2_INITIAL_PRESS = 2000
ZLL_SWITCH_BUTTON_3_INITIAL_PRESS = 3000
ZLL_SWITCH_BUTTON_4_INITIAL_PRESS = 4000

ZLL_SWITCH_BUTTON_1_HOLD = 1001
ZLL_SWITCH_BUTTON_2_HOLD = 2001
ZLL_SWITCH_BUTTON_3_HOLD = 3001
ZLL_SWITCH_BUTTON_4_HOLD = 4001

ZLL_SWITCH_BUTTON_1_SHORT_RELEASED = 1002
ZLL_SWITCH_BUTTON_2_SHORT_RELEASED = 2002
ZLL_SWITCH_BUTTON_3_SHORT_RELEASED = 3002
ZLL_SWITCH_BUTTON_4_SHORT_RELEASED = 4002

ZLL_SWITCH_BUTTON_1_LONG_RELEASED = 1003
ZLL_SWITCH_BUTTON_2_LONG_RELEASED = 2003
ZLL_SWITCH_BUTTON_3_LONG_RELEASED = 3003
ZLL_SWITCH_BUTTON_4_LONG_RELEASED = 4003

EVENT_BUTTON = "button"
EVENT_LIGHTLEVEL = "light_level"
EVENT_MOTION = "motion"
EVENT_POWER = "device_power"
EVENT_TEMPERATURE = "temperature"


class Sensors(APIItems):
    """Represents Hue Sensors.

    https://developers.meethue.com/documentation/sensors-api
    """

    def __init__(self, logger, raw, request):
        super().__init__(logger, raw, request, "sensors", create_sensor)


class GenericSensor:
    """Represents the base Hue sensor."""

    ITEM_TYPE = "sensors"

    def __init__(self, id, raw, request):
        self.id = id
        self.raw = raw
        self._request = request
        self.last_event = None

    @property
    def name(self):
        return self.raw["name"]

    @property
    def type(self):
        return self.raw["type"]

    @property
    def modelid(self):
        return self.raw["modelid"]

    @property
    def manufacturername(self):
        return self.raw["manufacturername"]

    @property
    def productname(self):
        return self.raw.get("productname")

    @property
    def uniqueid(self):
        return self.raw.get("uniqueid")

    @property
    def swversion(self):
        return self.raw.get("swversion")

    @property
    def state(self):
        return self.raw["state"]

    @property
    def config(self):
        return self.raw["config"]


class GenericCLIPSensor(GenericSensor):
    @property
    def battery(self):
        return self.raw["state"].get("battery")

    @property
    def lastupdated(self):
        return self.raw["state"]["lastupdated"]

    @property
    def on(self):
        return self.raw["config"]["on"]

    @property
    def reachable(self):
        return self.raw["config"]["reachable"]

    @property
    def url(self):
        return self.raw["config"].get("url")

    async def set_config(self, config):
        """Change config of a CLIP sensor."""
        await self._request("put", f"sensors/{self.id}/config", json=config)

    async def set_state(self, state):
        """Change state of a CLIP sensor."""
        await self._request("put", f"sensors/{self.id}/state", json=state)


class GenericZLLSensor(GenericSensor):
    @property
    def battery(self):
        return self.raw["state"].get("battery", self.raw["config"].get("battery"))

    @property
    def lastupdated(self):
        return self.raw["state"].get("lastupdated")

    @property
    def on(self):
        return self.raw["config"]["on"]

    @property
    def reachable(self):
        return self.raw["config"]["reachable"]


class GenericSwitchSensor:
    @property
    def buttonevent(self):
        return self.raw["state"]["buttonevent"]

    @property
    def inputs(self):
        return self.raw.get("capabilities", {}).get("inputs")

    async def set_config(self, on=None):
        """Change config of a Switch sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class DaylightSensor(GenericSensor):
    @property
    def configured(self):
        return self.raw["config"]["configured"]

    @property
    def daylight(self):
        return self.raw["state"]["daylight"]

    @property
    def on(self):
        return self.raw["config"]["on"]

    @property
    def sunriseoffset(self):
        return self.raw["config"]["sunriseoffset"]

    @property
    def sunsetoffset(self):
        return self.raw["config"]["sunsetoffset"]

    async def set_config(
        self, on=None, long=None, lat=None, sunriseoffset=None, sunsetoffset=None
    ):
        """Change config of a Daylight sensor."""
        data = {
            key: value
            for key, value in {
                "on": on,
                "long": long,
                "lat": lat,
                "sunriseoffset": sunriseoffset,
                "sunsetoffset": sunsetoffset,
            }.items()
            if value is not None
        }

        await self._request("put", f"sensors/{self.id}/config", json=data)


class GeofenceSensor(GenericSensor):
    @property
    def on(self):
        return self.raw["config"]["on"]

    @property
    def presence(self):
        return self.raw["state"]["presence"]

    @property
    def reachable(self):
        return self.raw["config"]["reachable"]

    async def set_config(self, on=None):
        """Change config of the Geofence sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class CLIPPresenceSensor(GenericCLIPSensor):
    @property
    def presence(self):
        return self.raw["state"]["presence"]

    async def set_config(self, on=None):
        """Change config of a CLIP Presence sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class ZLLPresenceSensor(GenericZLLSensor):
    @property
    def presence(self):
        return self.raw["state"]["presence"]

    async def set_config(self, on=None, sensitivity=None, sensitivitymax=None):
        """Change config of a ZLL Presence sensor."""
        data = {
            key: value
            for key, value in {
                "on": on,
                "sensitivity": sensitivity,
                "sensitivitymax": sensitivitymax,
            }.items()
            if value is not None
        }

        await self._request("put", f"sensors/{self.id}/config", json=data)


class ZLLRotarySensor(GenericZLLSensor):
    @property
    def rotaryevent(self):
        return self.raw["state"]["rotaryevent"]

    @property
    def expectedrotation(self):
        return self.raw["state"]["expectedrotation"]

    @property
    def expectedeventduration(self):
        return self.raw["state"]["expectedeventduration"]

    async def set_config(self, on=None):
        """Change config of a ZLL Rotary sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class CLIPSwitchSensor(GenericCLIPSensor):
    @property
    def buttonevent(self):
        return self.raw["state"]["buttonevent"]

    async def set_config(self, on=None):
        """Change config of a CLIP Switch sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class ZGPSwitchSensor(GenericSensor, GenericSwitchSensor):
    @property
    def lastupdated(self):
        return self.raw["state"].get("lastupdated")

    @property
    def on(self):
        return self.raw["config"]["on"]


class ZLLSwitchSensor(GenericZLLSensor, GenericSwitchSensor):
    pass


class CLIPLightLevelSensor(GenericCLIPSensor):
    @property
    def dark(self):
        return self.raw["state"]["dark"]

    @property
    def daylight(self):
        return self.raw["state"]["daylight"]

    @property
    def lightlevel(self):
        return self.raw["state"]["lightlevel"]

    @property
    def tholddark(self):
        return self.raw["config"]["tholddark"]

    @property
    def tholdoffset(self):
        return self.raw["config"]["tholdoffset"]

    async def set_config(self, on=None, tholddark=None, tholdoffset=None):
        """Change config of a CLIP LightLevel sensor."""
        data = {
            key: value
            for key, value in {
                "on": on,
                "tholddark": tholddark,
                "tholdoffset": tholdoffset,
            }.items()
            if value is not None
        }

        await self._request("put", f"sensors/{self.id}/config", json=data)


class ZLLLightLevelSensor(GenericZLLSensor):
    @property
    def dark(self):
        return self.raw["state"]["dark"]

    @property
    def daylight(self):
        return self.raw["state"]["daylight"]

    @property
    def lightlevel(self):
        return self.raw["state"]["lightlevel"]

    @property
    def tholddark(self):
        return self.raw["config"]["tholddark"]

    @property
    def tholdoffset(self):
        return self.raw["config"]["tholdoffset"]

    async def set_config(self, on=None, tholddark=None, tholdoffset=None):
        """Change config of a ZLL LightLevel sensor."""
        data = {
            key: value
            for key, value in {
                "on": on,
                "tholddark": tholddark,
                "tholdoffset": tholdoffset,
            }.items()
            if value is not None
        }

        await self._request("put", f"sensors/{self.id}/config", json=data)


class CLIPTemperatureSensor(GenericCLIPSensor):
    @property
    def temperature(self):
        return self.raw["state"]["temperature"]

    async def set_config(self, on=None):
        """Change config of a CLIP Temperature sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class ZLLTemperatureSensor(GenericZLLSensor):
    @property
    def temperature(self):
        return self.raw["state"]["temperature"]

    async def set_config(self, on=None):
        """Change config of a ZLL Temperature sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class CLIPGenericFlagSensor(GenericCLIPSensor):
    @property
    def flag(self):
        return self.raw["state"]["flag"]

    async def set_config(self, on=None):
        """Change config of a CLIP Generic Flag sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class CLIPGenericStatusSensor(GenericCLIPSensor):
    @property
    def status(self):
        return self.raw["state"]["status"]

    async def set_config(self, on=None):
        """Change config of a CLIP Generic Status sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class CLIPHumiditySensor(GenericCLIPSensor):
    @property
    def humidity(self):
        return self.raw["state"]["humidity"]

    async def set_config(self, on=None):
        """Change config of a CLIP Humidity sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


class CLIPOpenCloseSensor(GenericCLIPSensor):
    @property
    def open(self):
        return self.raw["state"]["open"]

    async def set_config(self, on=None):
        """Change config of a CLIP Open Close sensor."""
        data = {} if on is None else {"on": on}

        await self._request("put", f"sensors/{self.id}/config", json=data)


def create_sensor(id, raw, request):
    type = raw["type"]

    if type == TYPE_DAYLIGHT:
        return DaylightSensor(id, raw, request)

    elif type == TYPE_CLIP_GENERICFLAG:
        return CLIPGenericFlagSensor(id, raw, request)
    elif type == TYPE_CLIP_GENERICSTATUS:
        return CLIPGenericStatusSensor(id, raw, request)
    elif type == TYPE_CLIP_HUMIDITY:
        return CLIPHumiditySensor(id, raw, request)
    elif type == TYPE_CLIP_LIGHTLEVEL:
        return CLIPLightLevelSensor(id, raw, request)
    elif type == TYPE_CLIP_OPENCLOSE:
        return CLIPOpenCloseSensor(id, raw, request)
    elif type == TYPE_CLIP_PRESENCE:
        return CLIPPresenceSensor(id, raw, request)
    elif type == TYPE_CLIP_SWITCH:
        return CLIPSwitchSensor(id, raw, request)
    elif type == TYPE_CLIP_TEMPERATURE:
        return CLIPTemperatureSensor(id, raw, request)

    elif type == TYPE_GEOFENCE:
        return GeofenceSensor(id, raw, request)

    elif type == TYPE_ZGP_SWITCH:
        return ZGPSwitchSensor(id, raw, request)

    elif type == TYPE_ZLL_LIGHTLEVEL:
        return ZLLLightLevelSensor(id, raw, request)
    elif type == TYPE_ZLL_PRESENCE:
        return ZLLPresenceSensor(id, raw, request)
    elif type == TYPE_ZLL_ROTARY:
        return ZLLRotarySensor(id, raw, request)
    elif type == TYPE_ZLL_SWITCH:
        return ZLLSwitchSensor(id, raw, request)
    elif type == TYPE_ZLL_TEMPERATURE:
        return ZLLTemperatureSensor(id, raw, request)

    else:
        return GenericSensor(id, raw, request)



"""AIOHue implementation for HueBridge over V1/legacy api."""

from __future__ import annotations

import asyncio
import logging
from types import TracebackType

import aiohttp
from asyncio_throttle import Throttler

from aiohue.errors import BridgeBusy, Unauthorized, raise_from_error

from .config import Config
from .groups import Groups
from .lights import Lights
from .scenes import Scenes
from .sensors import Sensors

# how many times do we retry on a 503 or 429 (bridge overload/rate limit)
MAX_RETRIES = 25
THROTTLE_CONCURRENT_REQUESTS = 1  # how many concurrent requests to the bridge
THROTTLE_TIMESPAN = 0.25  # timespan/period (in seconds) for the rate limiting


class HueBridgeV1:
    """Control a Hue bridge with legacy/V1 API.."""

    def __init__(
        self,
        host: str,
        app_key: str,
        websession: aiohttp.ClientSession | None = None,
    ) -> None:
        """
        Initialize the Bridge instance.

        Parameters:
            `host`: the hostname or IP-address of the bridge as string.
            `app_key`: provide the hue appkey/username for authentication.
            `websession`: optionally provide a aiohttp ClientSession.
        """
        self._host = host
        self._app_key = app_key
        self._websession = websession
        self._websession_provided = websession is not None

        self.logger = logging.getLogger(f"{__package__}[{host}]")
        # all api controllers
        self._config = None
        self._devices = None
        self._lights = None
        self._scenes = None
        self._groups = None
        self._sensors = None
        # Setup the Throttler/rate limiter for requests to the bridge.
        self._throttler = Throttler(
            rate_limit=THROTTLE_CONCURRENT_REQUESTS, period=THROTTLE_TIMESPAN
        )

    @property
    def bridge_id(self) -> str | None:
        """Return the ID of the bridge we're currently connected to."""
        return self._config.bridge_id if self._config else None

    @property
    def host(self) -> str:
        """Return the hostname of the bridge."""
        return self._host

    @property
    def config(self) -> Config | None:
        """Get the bridge config."""
        return self._config

    @property
    def lights(self) -> Lights | None:
        """Get the light resources."""
        return self._lights

    @property
    def scenes(self) -> Scenes | None:
        """Get the scene resources."""
        return self._scenes

    @property
    def groups(self) -> Groups | None:
        """Get the group resources."""
        return self._groups

    @property
    def sensors(self) -> Sensors | None:
        """Get the sensor resources."""
        return self._sensors

    async def initialize(self):
        """Initialize the connection to the bridge and fetch all data."""
        result = await self.request("get", "")
        self._config = Config(result.pop("config"), self.request)
        self._groups = Groups(self.logger, result.pop("groups"), self.request)
        self._lights = Lights(self.logger, result.pop("lights"), self.request)
        if "scenes" in result:
            self._scenes = Scenes(self.logger, result.pop("scenes"), self.request)
        if "sensors" in result:
            self._sensors = Sensors(self.logger, result.pop("sensors"), self.request)
        self.logger.debug("Unused result: %s", result)

    async def close(self) -> None:
        """Close connection and cleanup."""
        if not self._websession_provided:
            await self._websession.close()
        self.logger.info("Connection to bridge closed.")

    async def request(self, method, endpoint, json=None):
        """Make request on the api and return response data."""
        if self._websession is None:
            self._websession = aiohttp.ClientSession()

        # Old bridges and (most) emulators only use `http`
        url = f"http://{self.host}/api/{self._app_key}/{endpoint}"
        # The bridge will rate limit if we send more requests than about 2-5 per second
        # we guard ourselves from hitting the rate limit by using a throttler
        # but others apps/services are hitting the Hue bridge too so we still
        # might hit the rate limit/overload at some point so we also retry if this happens.
        retries = 0

        while retries < MAX_RETRIES:
            retries += 1

            if retries > 1:
                retry_wait = 0.25 * retries
                self.logger.debug(
                    "Got 503 or 429 error from Hue bridge, retry request in %s seconds",
                    retry_wait,
                )
                await asyncio.sleep(retry_wait)

            async with self._websession.request(method, url, json=json) as resp:
                # 503 means the service is temporarily unavailable, back off a bit.
                if resp.status == 503:
                    continue
                # 429 means the bridge is rate limiting/overloaded, we should back off a bit.
                if resp.status == 429:
                    continue
                if resp.status == 403:
                    raise Unauthorized
                # raise on all other error status codes
                resp.raise_for_status()
                data = await resp.json()
                _raise_on_error(data)
                return data

        raise BridgeBusy(
            f"{retries} requests to the bridge failed, "
            "its probably overloaded. Giving up."
        )

    async def __aenter__(self) -> HueBridgeV1:
        """Return Context manager."""
        await self.initialize()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> bool | None:
        """Exit context manager."""
        await self.close()
        if exc_val:
            raise exc_val
        return exc_type


def _raise_on_error(data):
    """Check response for error message."""
    if isinstance(data, list):
        data = data[0]

    if isinstance(data, dict) and "error" in data:
        raise_from_error(data["error"])



"""
Model(s) for grouped_light resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light
"""

from dataclasses import dataclass

from .feature import (
    AlertFeature,
    AlertFeaturePut,
    ColorFeaturePut,
    ColorTemperatureDeltaFeaturePut,
    ColorTemperatureFeaturePut,
    DimmingDeltaFeaturePut,
    DimmingFeatureBase,
    DynamicsFeaturePut,
    OnFeature,
    SignalingFeature,
    SignalingFeaturePut,
)
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class GroupedLight:
    """
    Represent a (full) GroupedLight object when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light_get
    """

    id: str
    owner: ResourceIdentifier
    id_v1: str | None = None
    # on: Joined on control & aggregated on state.
    # “on” is true if any light in the group is on.
    on: OnFeature | None = None
    # dimming: Joined dimming control.
    # “dimming.brightness” contains average brightness of group containing turned-on lights only.
    dimming: DimmingFeatureBase | None = None
    alert: AlertFeature | None = None  # Joined alert control
    signaling: SignalingFeature | None = None
    type: ResourceTypes = ResourceTypes.GROUPED_LIGHT


@dataclass
class GroupedLightPut:
    """
    Represent a GroupedLight model when sending a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_light__id__put
    """

    on: OnFeature | None = None
    dimming: DimmingFeatureBase | None = None
    dimming_delta: DimmingDeltaFeaturePut | None = None
    color_temperature: ColorTemperatureFeaturePut | None = None
    color_temperature_delta: ColorTemperatureDeltaFeaturePut | None = None
    color: ColorFeaturePut | None = None
    dynamics: DynamicsFeaturePut | None = None
    alert: AlertFeaturePut | None = None
    signaling: SignalingFeaturePut | None = None



"""Example/test script for (stress) testing multiple requests to the bridge."""

import argparse
import asyncio
import contextlib
import logging
import time

from aiohue import HueBridgeV2

parser = argparse.ArgumentParser(description="AIOHue Example")
parser.add_argument("host", help="hostname of Hue bridge")
parser.add_argument("appkey", help="appkey for Hue bridge")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main():
    """Run Main execution."""
    if args.debug:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(asctime)-15s %(levelname)-5s %(name)s -- %(message)s",
        )
        asyncio.get_running_loop().set_debug(True)

    async with HueBridgeV2(args.host, args.appkey) as bridge:
        print("Connected to bridge: ", bridge.bridge_id)
        print(bridge.config.bridge_device)

        # pick a random light
        light = bridge.lights.items[0]
        print(f"Sending 100 requests to bridge for {light.name}...")

        async def toggle_light():
            await bridge.lights.turn_on(light.id)
            await bridge.lights.turn_off(light.id)

        before = time.time()
        await asyncio.gather(*[toggle_light() for i in range(50)])
        after = time.time()
        print(f"Completed in {after - before} seconds...")


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Control a Philips Hue bridge with V2 API."""

import asyncio
import logging
import time
from collections.abc import Callable, Generator
from contextlib import asynccontextmanager
from types import TracebackType
from typing import Any
from uuid import uuid4

import aiohttp
from aiohttp import ClientResponse

from aiohue.errors import BridgeBusy, Unauthorized, raise_from_error

from .controllers.config import ConfigController
from .controllers.devices import DevicesController
from .controllers.events import EventCallBackType, EventStream, EventType
from .controllers.groups import GroupsController
from .controllers.lights import LightsController
from .controllers.scenes import ScenesController
from .controllers.sensors import SensorsController

# how many times do we retry on a 503 or 429 (bridge overload/rate limit)
MAX_RETRIES = 25


class HueBridgeV2:
    """Control a Philips Hue bridge with V2 API."""

    _websession: aiohttp.ClientSession | None = None

    def __init__(
        self,
        host: str,
        app_key: str,
    ) -> None:
        """
        Initialize the Bridge instance.

        Parameters:
            `host`: the hostname or IP-address of the bridge as string.
            `app_key`: provide the hue appkey/username for authentication.
        """
        self._host = host
        self._app_key = app_key
        self.logger = logging.getLogger(f"{__package__}[{host}]")
        self._events = EventStream(self)
        # all resource controllers
        self._config = ConfigController(self)
        self._devices = DevicesController(self)
        self._lights = LightsController(self)
        self._scenes = ScenesController(self)
        self._groups = GroupsController(self)
        self._sensors = SensorsController(self)
        self._disconnect_timestamp = 0

    @property
    def bridge_id(self) -> str | None:
        """Return the ID of the bridge we're currently connected to."""
        return self._config.bridge_id

    @property
    def host(self) -> str:
        """Return the hostname of the bridge."""
        return self._host

    @property
    def events(self) -> EventStream:
        """Return the EventStream controller."""
        return self._events

    @property
    def config(self) -> ConfigController:
        """Get the Config Controller with config-like resources."""
        return self._config

    @property
    def devices(self) -> DevicesController:
        """Get the Devices Controller for managing all device resources."""
        return self._devices

    @property
    def lights(self) -> LightsController:
        """Get the Lights Controller for managing all light resources."""
        return self._lights

    @property
    def scenes(self) -> ScenesController:
        """Get the Scenes Controller for managing all scene resources."""
        return self._scenes

    @property
    def groups(self) -> GroupsController:
        """Get the Groups Controller for managing all group resources."""
        return self._groups

    @property
    def sensors(self) -> SensorsController:
        """Get the Sensors Controller for managing all sensor resources."""
        return self._sensors

    async def initialize(self) -> None:
        """Initialize the connection to the bridge and fetch all data."""
        # Initialize all HUE resource controllers
        # fetch complete full state once and distribute to controllers
        await self.fetch_full_state()
        # start event listener
        await self._events.initialize()
        # subscribe to reconnect event
        self._events.subscribe(
            self._handle_connect_event, (EventType.RECONNECTED, EventType.DISCONNECTED)
        )

    async def close(self) -> None:
        """Close connection and cleanup."""
        await self.events.stop()
        if self._websession:
            await self._websession.close()
        self.logger.info("Connection to bridge closed.")

    def subscribe(
        self,
        callback: EventCallBackType,
    ) -> Callable:
        """
        Subscribe to status changes for all resources.

        Returns:
            function to unsubscribe.
        """
        unsubscribes = [
            self.config.subscribe(callback),
            self.devices.subscribe(callback),
            self.groups.subscribe(callback),
            self.lights.subscribe(callback),
            self.scenes.subscribe(callback),
            self.sensors.subscribe(callback),
        ]

        def unsubscribe():
            for unsub in unsubscribes:
                unsub()

        return unsubscribe

    async def request(self, method: str, path: str, **kwargs) -> dict | list[dict]:
        """Make request on the api and return response data."""
        # The bridge will deny more than 3 requests at the same time with a 429 error.
        # We guard ourselves from hitting this error by limiting the TCP Connector for aiohttp
        # but other apps/services are hitting the Hue bridge too so we still
        # might hit the rate limit/overload at some point
        # so we have some retry logic if this happens.
        retries = 0

        while retries < MAX_RETRIES:
            retries += 1

            if retries > 1:
                retry_wait = 0.25 * retries
                self.logger.debug(
                    "Got 503 or 429 error from Hue bridge, retry request in %s seconds",
                    retry_wait,
                )
                await asyncio.sleep(retry_wait)

            async with self.create_request(method, path, **kwargs) as resp:
                # 503 means the service is temporarily unavailable, back off a bit.
                if resp.status == 503:
                    continue
                # 429 means the bridge is rate limiting/overloaded, we should back off a bit.
                if resp.status == 429:
                    continue
                if resp.status == 403:
                    raise Unauthorized
                # raise on all other error status codes
                resp.raise_for_status()
                result = await resp.json()
                if result.get("errors"):
                    raise_from_error(result["errors"][0])
                return result["data"]

        raise BridgeBusy(
            f"{retries} requests to the bridge failed, "
            "its probably overloaded. Giving up."
        )

    @asynccontextmanager
    async def create_request(
        self, method: str, path: str, **kwargs
    ) -> Generator[ClientResponse, None, None]:
        """
        Make a request to any path with V2 request method (auth in header).

        Returns a generator with aiohttp ClientResponse.
        """
        if self._websession is None:
            connector = aiohttp.TCPConnector(
                limit_per_host=3,
            )
            self._websession = aiohttp.ClientSession(connector=connector)

        url = f"https://{self._host}/{path}"

        kwargs["ssl"] = False

        if "headers" not in kwargs:
            kwargs["headers"] = {}

        kwargs["headers"]["hue-application-key"] = self._app_key

        async with self._websession.request(method, url, **kwargs) as res:
            yield res

    async def __aenter__(self) -> "HueBridgeV2":
        """Return Context manager."""
        await self.initialize()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> bool | None:
        """Exit context manager."""
        await self.close()
        if exc_val:
            raise exc_val
        return exc_type

    async def _handle_connect_event(
        self,
        event_type: EventType,
        item: Any = None,  # noqa: ARG002
    ) -> None:
        """Handle (disconnect) event from the EventStream."""
        # pylint: disable=unused-argument
        if event_type == EventType.DISCONNECTED:
            # If we receive a disconnect event, we store the timestamp
            self._disconnect_timestamp = time.time()
        elif event_type == EventType.RECONNECTED:  # noqa: SIM102
            # if the time between the disconnect and reconnect is more than 1 minute,
            # we fetch the full state.
            if (time.time() - self._disconnect_timestamp) > 60:
                await self.fetch_full_state()

    async def fetch_full_state(self) -> None:
        """Fetch state on all controllers."""
        full_state = await self.request("get", "clip/v2/resource")
        await asyncio.gather(
            self._config.initialize(full_state),
            self._devices.initialize(full_state),
            self._lights.initialize(full_state),
            self._scenes.initialize(full_state),
            self._sensors.initialize(full_state),
            self._groups.initialize(full_state),
        )

    async def get_diagnostics(self) -> dict[str, Any]:
        """Return a dict with diagnostic information for debugging and support purposes."""
        result = {}
        subst_ids = {}

        def _replace_id(item: dict):
            """Replace ID with a random one for privacy reasons."""
            id_key = "id" if "id" in item else "rid"
            org_item_id = item[id_key]
            if org_item_id not in subst_ids:
                subst_ids[org_item_id] = f"redacted-{uuid4()}"
            new_id = subst_ids[org_item_id]
            item[id_key] = new_id

        # add full state to result
        full_state = await self.request("get", "clip/v2/resource")
        for item in full_state:
            _replace_id(item)
        result["full_state"] = full_state

        # add last event messages to result
        last_events = []
        for item in self._events.last_events:
            if "id" in item or "rid" in item:
                # copy the object so we're not modifying the original
                item = {**item}  # noqa: PLW2901
                _replace_id(item)
            last_events.append(item)
        result["events"] = last_events

        return result



"""
Model(s) for motion_area_configuration resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_motion_area_configuration
"""

from dataclasses import dataclass, field
from enum import Enum

from .resource import ResourceIdentifier, ResourceTypes


class MotionAreaHealth(Enum):
    """Enum with possible health status values."""

    HEALTHY = "healthy"
    DEGRADED = "degraded"
    RECOVERING = "recovering"
    UNRECOVERABLE = "unrecoverable"
    NOT_RUNNING = "not_running"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, value: object):  # noqa: ARG003
        """Set default enum member if an unknown value is provided."""
        return MotionAreaHealth.UNKNOWN


@dataclass
class MotionAreaStatus:
    """
    Represent health status information.

    Used by `motion_area_configuration` resources.
    """

    health: MotionAreaHealth = MotionAreaHealth.UNKNOWN


@dataclass
class MotionAreaParticipant:
    """
    Represent a participant in motion area configuration.

    Used by `motion_area_configuration` resources.
    """

    resource: ResourceIdentifier
    status: MotionAreaStatus = field(default_factory=MotionAreaStatus)


@dataclass
class MotionAreaConfiguration:
    """
    Represent a (full) `MotionAreaConfiguration` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_motion_area_configuration_get
    """

    id: str
    name: str
    group: ResourceIdentifier
    participants: list[MotionAreaParticipant]
    services: list[ResourceIdentifier]
    health: MotionAreaHealth
    enabled: bool

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.MOTION_AREA_CONFIGURATION


@dataclass
class MotionAreaConfigurationPut:
    """
    MotionAreaConfiguration resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_motion_area_configuration__id__put
    """

    name: str | None = None
    enabled: bool | None = None



from __future__ import annotations

from collections import namedtuple
from collections.abc import Coroutine
from logging import Logger
from typing import Any

from .api import APIItems

# Represents a CIE 1931 XY coordinate pair.
XYPoint = namedtuple("XYPoint", ["x", "y"])


# Represents the Gamut of a light.
GamutType = namedtuple("GamutType", ["red", "green", "blue"])


class Lights(APIItems):
    """
    Represents Hue Lights.

    https://developers.meethue.com/documentation/lights-api
    """

    def __init__(self, logger: Logger, raw: dict[str, Any], request: Coroutine) -> None:
        """Initialize instance."""
        super().__init__(logger, raw, request, "lights", Light)


class Light:
    """Represents a Hue light."""

    ITEM_TYPE = "lights"

    def __init__(self, id: str, raw: dict[str, Any], request: Coroutine) -> None:
        """Initialize instance."""
        self.id = id
        self.raw = raw
        self._request = request

    @property
    def uniqueid(self):
        return self.raw["uniqueid"]

    @property
    def manufacturername(self):
        return self.raw["manufacturername"]

    @property
    def modelid(self):
        return self.raw["modelid"]

    @property
    def productname(self):
        # productname added in Bridge API 1.24 (published 03/05/2018)
        return self.raw.get("productname")

    @property
    def name(self):
        return self.raw["name"]

    @property
    def state(self):
        return self.raw["state"]

    @property
    def type(self):
        return self.raw["type"]

    @property
    def swversion(self):
        """Software version of the light."""
        return self.raw["swversion"]

    @property
    def swupdatestate(self):
        """Software update state of the light."""
        return self.raw.get("swupdate", {}).get("state")

    @property
    def controlcapabilities(self):
        """Capabilities that the light has to control it."""
        return self.raw.get("capabilities", {}).get("control", {})

    @property
    def colorgamuttype(self):
        """The color gamut type of the light."""
        light_spec = self.controlcapabilities
        return light_spec.get("colorgamuttype", "None")

    @property
    def colorgamut(self):
        """The color gamut information of the light."""
        try:
            light_spec = self.controlcapabilities
            gtup = tuple([XYPoint(*x) for x in light_spec["colorgamut"]])
            color_gamut = GamutType(*gtup)
        except KeyError:
            color_gamut = None

        return color_gamut

    async def set_state(
        self,
        on=None,
        bri=None,
        hue=None,
        sat=None,
        xy=None,
        ct=None,
        alert=None,
        effect=None,
        transitiontime=None,
        bri_inc=None,
        sat_inc=None,
        hue_inc=None,
        ct_inc=None,
        xy_inc=None,
    ):
        """Change state of a light."""
        data = {
            key: value
            for key, value in {
                "on": on,
                "bri": bri,
                "hue": hue,
                "sat": sat,
                "xy": xy,
                "ct": ct,
                "alert": alert,
                "effect": effect,
                "transitiontime": transitiontime,
                "bri_inc": bri_inc,
                "sat_inc": sat_inc,
                "hue_inc": hue_inc,
                "ct_inc": ct_inc,
                "xy_inc": xy_inc,
            }.items()
            if value is not None
        }

        await self._request("put", f"lights/{self.id}/state", json=data)



"""
Model(s) for zigbee_connectivity resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zigbee_connectivity
"""

from dataclasses import dataclass
from enum import Enum

from .resource import ResourceIdentifier, ResourceTypes


class ConnectivityServiceStatus(Enum):
    """
    Enum with possible ConnectivityService statuses.

    Connected if device has been recently been available.
    When indicating connectivity issues the device is powered off or has network
    issues When indicating unidirectional incoming the device only talks to bridge
    pending_discovery when device is expected to be discovered (added to the
    network) soon
    """

    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    CONNECTIVITY_ISSUE = "connectivity_issue"
    UNIDIRECTIONAL_INCOMING = "unidirectional_incoming"
    PENDING_DISCOVERY = "pending_discovery"


@dataclass
class ZigbeeConnectivity:
    """
    Represent a (full) `ZigbeeConnectivity` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zigbee_connectivity_get
    """

    id: str
    owner: ResourceIdentifier
    status: ConnectivityServiceStatus
    mac_address: str
    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.ZIGBEE_CONNECTIVITY



"""
Model(s) for temperature resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_temperature
"""

from dataclasses import dataclass
from datetime import datetime

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class TemperatureReport:
    """
    Represent TemperatureReport as retrieved from api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_temperature_get
    """

    changed: datetime
    temperature: float


@dataclass
class TemperatureSensingFeature:
    """Represent TemperatureFeature."""

    temperature_report: TemperatureReport | None = None
    temperature: float | None = None  # deprecated
    temperature_valid: bool | None = None  # deprecated

    @property
    def value(self) -> float | None:
        """Return the actual/current value."""
        # prefer new style attribute (not available on older firmware versions)
        if self.temperature_report is not None:
            return self.temperature_report.temperature
        return self.temperature


@dataclass
class Temperature:
    """
    Represent a (full) `Temperature` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_temperature_get
    """

    id: str
    owner: ResourceIdentifier
    enabled: bool
    temperature: TemperatureSensingFeature

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.TEMPERATURE


@dataclass
class TemperaturePut:
    """
    Temperature resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_temperature__id__put
    """

    enabled: bool | None = None



"""
Model(s) for Light resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light
"""

from dataclasses import dataclass
from enum import Enum

from .feature import (
    AlertFeature,
    AlertFeaturePut,
    ColorFeature,
    ColorFeatureBase,
    ColorTemperatureDeltaFeaturePut,
    ColorTemperatureFeature,
    ColorTemperatureFeatureBase,
    DimmingDeltaFeaturePut,
    DimmingFeature,
    DimmingFeatureBase,
    DynamicsFeature,
    DynamicsFeaturePut,
    DynamicStatus,
    EffectsFeature,
    EffectsFeaturePut,
    GradientFeature,
    GradientFeatureBase,
    OnFeature,
    PowerUpFeature,
    PowerUpFeaturePut,
    SignalingFeature,
    SignalingFeaturePut,
    TimedEffectsFeature,
    TimedEffectsFeaturePut,
)
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class LightMetaData:
    """
    Represent LightMetaData object as used by the Hue api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    # Light archetype Deprecated: use archetype on device level
    archetype: str | None
    name: str


class LightMode(Enum):
    """
    Enum with all possible LightModes.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    NORMAL = "normal"
    STREAMING = "streaming"


@dataclass
class Light:
    """
    Represent a (full) `Light` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_get
    """

    # pylint: disable=too-many-instance-attributes

    id: str
    owner: ResourceIdentifier
    on: OnFeature
    mode: LightMode

    id_v1: str | None = None
    dimming: DimmingFeature | None = None
    color_temperature: ColorTemperatureFeature | None = None
    color: ColorFeature | None = None
    dynamics: DynamicsFeature | None = None
    alert: AlertFeature | None = None
    signaling: SignalingFeature | None = None
    gradient: GradientFeature | None = None
    effects: EffectsFeature | None = None
    timed_effects: TimedEffectsFeature | None = None
    powerup: PowerUpFeature | None = None

    type: ResourceTypes = ResourceTypes.LIGHT

    @property
    def supports_dimming(self) -> bool:
        """Return if this light supports brightness control."""
        return self.dimming is not None

    @property
    def supports_color(self) -> bool:
        """Return if this light supports color control."""
        return self.color is not None

    @property
    def supports_color_temperature(self) -> bool:
        """Return if this light supports color_temperature control."""
        return self.color_temperature is not None

    @property
    def is_on(self) -> bool:
        """Return bool if light is currently powered on."""
        if self.on is not None:
            return self.on.on
        return False

    @property
    def brightness(self) -> float:
        """Return current brightness of light."""
        if self.dimming is not None:
            return self.dimming.brightness
        return 100.0 if self.is_on else 0.0

    @property
    def is_dynamic(self) -> bool:
        """Return bool if light is currently dynamically changing colors from a scene."""
        if self.dynamics is not None:
            return self.dynamics.status == DynamicStatus.DYNAMIC_PALETTE
        return False

    @property
    def entertainment_active(self) -> bool:
        """Return bool if this light is currently streaming in Entertainment mode."""
        return self.mode == LightMode.STREAMING


@dataclass
class LightPut:
    """
    Light resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light__id__put
    """

    on: OnFeature | None = None
    dimming: DimmingFeatureBase | None = None
    dimming_delta: DimmingDeltaFeaturePut | None = None
    color_temperature: ColorTemperatureFeatureBase | None = None
    color_temperature_delta: ColorTemperatureDeltaFeaturePut | None = None
    color: ColorFeatureBase | None = None
    dynamics: DynamicsFeaturePut | None = None
    alert: AlertFeaturePut | None = None
    gradient: GradientFeatureBase | None = None
    effects: EffectsFeaturePut | None = None
    timed_effects: TimedEffectsFeaturePut | None = None
    powerup: PowerUpFeaturePut | None = None
    signaling: SignalingFeaturePut | None = None



"""Handle connecting to the HUE Eventstream and distribute events."""

import asyncio
from collections import deque
from collections.abc import Callable
from enum import Enum
from inspect import iscoroutinefunction
import json
import random
import string
from typing import TYPE_CHECKING, NoReturn, TypedDict

from aiohttp import ClientTimeout
from aiohttp.client_exceptions import ClientError

from aiohue.errors import AiohueException, InvalidAPIVersion, InvalidEvent, Unauthorized
from aiohue.util import NoneType
from aiohue.v2.models.geofence_client import GeofenceClientPost, GeofenceClientPut
from aiohue.v2.models.resource import ResourceTypes

if TYPE_CHECKING:
    from aiohue.v2 import HueBridgeV2


CONNECTION_TIMEOUT = 90  # 90 seconds
KEEPALIVE_INTERVAL = 60  # every minute


class EventStreamStatus(Enum):
    """Status options of EventStream."""

    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class HueEvent(TypedDict):
    """Hue Event message as emitted by the EventStream."""

    id: str  # UUID
    creationtime: str
    type: str  # = EventType (add, update, delete)
    # data contains a list with (partial) resource objects
    # in case of add or update this is a full or partial resource object
    # in case of delete this will include only the
    # ResourceIndentifier (type and id) of the deleted object
    data: list[dict]


class EventType(Enum):
    """Enum with possible Events."""

    # resource events match those emitted by Hue eventstream
    RESOURCE_ADDED = "add"
    RESOURCE_UPDATED = "update"
    RESOURCE_DELETED = "delete"
    # connection events emitted by (this) events controller
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    RECONNECTED = "reconnected"


EventCallBackType = Callable[[EventType, dict | None], None]
EventSubscriptionType = tuple[
    EventCallBackType,
    "tuple[EventType] | None",
    "tuple[ResourceTypes] | None",
]


class EventStream:
    """Holds the connection to the HUE Clip EventStream."""

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._listeners = set()
        self._event_queue = asyncio.Queue()
        self._last_event_id = ""
        self._status = EventStreamStatus.DISCONNECTED
        self._bg_tasks: list[asyncio.Task] = []
        self._subscribers: list[EventSubscriptionType] = []
        self._logger = bridge.logger.getChild("events")
        self._event_history = deque(maxlen=25)

    @property
    def connected(self) -> bool:
        """Return bool if we're connected."""
        return self._status == EventStreamStatus.CONNECTED

    @property
    def status(self) -> bool:
        """Return connection status."""
        return self._status

    @property
    def last_events(self) -> list[dict]:
        """Return a list with the previous X messages."""
        return self._event_history

    async def initialize(self) -> None:
        """
        Start listening for events.

        Starts the connection to the Hue Eventstream and collect events.
        Connection will be auto-reconnected if it gets lost.
        """
        assert len(self._bg_tasks) == 0
        self._bg_tasks.append(asyncio.create_task(self.__event_reader()))
        self._bg_tasks.append(asyncio.create_task(self.__event_processor()))
        self._bg_tasks.append(asyncio.create_task(self.__keepalive_workaround()))

    async def stop(self) -> None:
        """Stop listening for events."""
        for task in self._bg_tasks:
            task.cancel()
        self._bg_tasks = []

    def subscribe(
        self,
        callback: EventCallBackType,
        event_filter: EventType | tuple[EventType] | None = None,
        resource_filter: ResourceTypes | tuple[ResourceTypes] | None = None,
    ) -> Callable:
        """
        Subscribe to events emitted by the Hue bridge for resources.

        Parameters:
            - `callback` - callback function to call when an event emits.
            - `event_filter` - Optionally provide an EventType as filter.
            - `resource_filter` - Optionally provide a ResourceType as filter.

        Returns:
            function to unsubscribe.
        """
        if not isinstance(event_filter, NoneType | tuple):
            event_filter = (event_filter,)
        if not isinstance(resource_filter, NoneType | tuple):
            resource_filter = (resource_filter,)
        subscription = (callback, event_filter, resource_filter)

        def unsubscribe():
            self._subscribers.remove(subscription)

        self._subscribers.append(subscription)
        return unsubscribe

    def emit(self, event_type: EventType, data: dict | None = None) -> None:
        """Emit event to all listeners."""
        for callback, event_filter, resource_filter in self._subscribers:
            if event_filter is not None and event_type not in event_filter:
                continue
            if (
                data is not None
                and resource_filter is not None
                and ResourceTypes(data.get("type")) not in resource_filter
            ):
                continue
            if iscoroutinefunction(callback):
                asyncio.create_task(callback(event_type, data))
            else:
                callback(event_type, data)

    async def __event_reader(self) -> None:
        """

        Read incoming SSE messages and put them in a Queue to be processed.

        Background task that keeps (re)connecting until stopped.

        https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation
        """
        self._status = EventStreamStatus.CONNECTING
        connect_attempts = 0
        headers = {
            "Accept": "text/event-stream",
            "Cache-Control": "no-cache",
            "Connection": "close",
        }

        while True:
            connect_attempts += 1
            if self._last_event_id:
                headers["last-event-id"] = self._last_event_id
            # Messages come in line by line, according to EventStream/SSE specs
            # we iterate over the incoming lines in the streamreader.
            try:
                async with self._bridge.create_request(
                    "get",
                    "eventstream/clip/v2",
                    timeout=ClientTimeout(total=0, sock_read=CONNECTION_TIMEOUT),
                    headers=headers,
                ) as resp:
                    # update status to connected once we reach this point
                    self._status = EventStreamStatus.CONNECTED
                    if connect_attempts == 1:
                        self.emit(EventType.CONNECTED)
                    else:
                        self.emit(EventType.RECONNECTED)
                    connect_attempts = 1  # reset on successful connect
                    self._logger.debug("Connected to EventStream")
                    # read over incoming messages line by line
                    async for line in resp.content:
                        # process the message
                        self.__parse_message(line)
            except (ClientError, asyncio.TimeoutError) as err:
                # pass expected connection errors because we will auto retry
                status = getattr(err, "status", None)
                if status == 404:
                    raise InvalidAPIVersion from err
                if status == 403:
                    raise Unauthorized from err
                if status:
                    # for debugging purpose only
                    self._logger.debug(err)
            except Exception as err:
                # for debugging purpose only
                self._logger.exception(err)
                raise err

            # if we reach this point, the connection was lost
            self.emit(EventType.DISCONNECTED)

            reconnect_wait = min(2 * connect_attempts, 600)
            self._logger.debug(
                "Disconnected from EventStream"
                " - Reconnect will be attempted in %s seconds",
                reconnect_wait,
            )
            # every 10 failed connect attempts log warning
            if connect_attempts % 10 == 0:
                self._logger.warning(
                    "%s Attempts to (re)connect to bridge failed"
                    " - This might be an indication of connection issues.",
                    connect_attempts,
                )
            self._status = EventStreamStatus.CONNECTING
            await asyncio.sleep(reconnect_wait)

    async def __event_processor(self) -> None:
        """Process incoming Hue events on the Queue and distribute those."""
        while True:
            event: HueEvent = await self._event_queue.get()
            # each clip event has array of updated/added/deleted objects in data property
            # we fire an event for each object that was added/updated/deleted
            for item in event["data"]:
                self.emit(EventType(event["type"]), item)

    def __parse_message(self, msg: bytes) -> None:
        """Parse a plain message string as received from EventStream."""
        try:
            line = msg.decode().strip()
            if not line or ":" not in line:
                return
            key, value = line.split(":", 1)
            if not key:
                return
            if key == "id":
                self._last_event_id = value
                return
            if key == "data":
                # events is array with multiple events
                events: list[HueEvent] = json.loads(value)
                for event in events:
                    if event.get("type") not in ["add", "update", "delete"]:
                        raise InvalidEvent(f"Received invalid event {event}")
                    self._event_queue.put_nowait(event)
                    self._event_history.append(event)
                return
            if key != "data":
                self._logger.debug("Received unexpected message: %s - %s", key, value)
        except Exception as exc:  # pylint: disable=broad-except
            self._logger.warning(
                "Unable to parse Event message: %s", line, exc_info=exc
            )

    async def __keepalive_workaround(self) -> NoReturn:
        """Send keepalive command to bridge, abusing geofence client."""
        # Oh yeah, this is a major hack and hopefully it will only be temporary ;-)
        # Signify forgot to implement some sort of periodic keep alive message on the EventBus
        # so we have no way to determine if the connection is still alive.
        # To workaround this, we create a geofence client (with status not at home)
        # on the bridge for aiohue which will have its name updated every minute
        # this will result in an event on the eventstream and thus a way to figure out
        # if its still alive. It's not very pretty but at least it works.
        # Now let's contact Signify if this can be solved.
        # Update 2022-12-12: Still no solution and apparently no real need for signify to fix this
        # https://developers.meethue.com/forum/t/several-issues-and-questions-about-the-v2-api/6586/35
        prefix = "aiohue_"

        while True:
            await asyncio.sleep(KEEPALIVE_INTERVAL)

            try:
                for client in self._bridge.sensors.geofence_client.items:
                    if client.name.startswith(prefix):
                        hass_client = client
                        break
                else:
                    await self._bridge.sensors.geofence_client.create(
                        GeofenceClientPost(name=prefix, is_at_home=False)
                    )
                    continue

                # simply updating the name of the geofence client will result in an event message
                # the eventstream timeout will detect if our keepalive message was not received
                random_str = "".join(
                    (random.choice(string.ascii_lowercase)) for x in range(10)
                )
                await self._bridge.sensors.geofence_client.update(
                    hass_client.id,
                    GeofenceClientPut(name=f"{prefix}{random_str}", is_at_home=False),
                )
            except (ClientError, asyncio.TimeoutError, AiohueException) as err:
                # might happen on connection error, we don't want the retry logic to bail out
                self._logger.debug("Error while sending keepalive: %s", str(err))



"""Controller holding and managing HUE group resources."""

import asyncio
from typing import TYPE_CHECKING, Union

from aiohue.v2.models.feature import (
    AlertEffectType,
    AlertFeaturePut,
    ColorFeaturePut,
    ColorPoint,
    ColorTemperatureFeaturePut,
    DeltaAction,
    DimmingFeaturePut,
    DimmingDeltaFeaturePut,
    DynamicsFeaturePut,
    OnFeature,
)
from aiohue.v2.models.grouped_light import GroupedLight, GroupedLightPut
from aiohue.v2.models.light import Light
from aiohue.v2.models.resource import ResourceTypes
from aiohue.v2.models.room import Room
from aiohue.v2.models.scene import Scene
from aiohue.v2.models.zone import Zone

from .base import BaseResourcesController, GroupedControllerBase

if TYPE_CHECKING:
    from aiohue.v2 import HueBridgeV2


class RoomController(BaseResourcesController[type[Room]]):
    """Controller holding and managing HUE resources of type `room`."""

    item_type = ResourceTypes.ROOM
    item_cls = Room
    allow_parser_error = True

    def get_scenes(self, id: str) -> list[Scene]:
        """Get all scenes for this room."""
        return [scene for scene in self._bridge.scenes if scene.group.rid == id]

    def get_lights(self, id: str) -> list[Light]:
        """Return all lights in given room."""
        if id not in self._items:
            return []
        result = []
        for dev_id in self._items[id].devices:
            if (dev := self._bridge.devices.get(dev_id)) is None:
                continue

            for light_id in dev.lights:
                if light := self._bridge.lights.get(light_id):
                    result.append(light)
        return result


class ZoneController(BaseResourcesController[type[Zone]]):
    """Controller holding and managing HUE resources of type `zone`."""

    item_type = ResourceTypes.ZONE
    item_cls = Zone
    allow_parser_error = True

    def get_scenes(self, id: str) -> list[Scene]:
        """Get all scenes for this room."""
        return [scene for scene in self._bridge.scenes if scene.group.rid == id]

    def get_lights(self, id: str) -> list[Light]:
        """Return all lights in given zone."""
        if id not in self._items:
            return []
        light_ids = {
            x.rid for x in self._items[id].children if x.rtype == ResourceTypes.LIGHT
        }
        return [x for x in self._bridge.lights if x.id in light_ids]


class GroupedLightController(BaseResourcesController[type[GroupedLight]]):
    """Controller holding and managing HUE resources of type `grouped_light`."""

    item_type = ResourceTypes.GROUPED_LIGHT
    item_cls = GroupedLight

    def get_zone(self, id: str) -> Room | Zone | None:
        """Get the zone or room connected to grouped light."""
        for group in self._bridge.groups:
            if group.type == ResourceTypes.GROUPED_LIGHT:
                continue
            if group.grouped_light == id:
                return group
        return None

    def get_lights(self, id: str) -> list[Light]:
        """Return lights of the connected room/zone."""
        # Note that this is just a convenience method for backwards compatibility
        if zone := self.get_zone(id):
            if zone.type == ResourceTypes.ROOM:
                return self._bridge.groups.room.get_lights(zone.id)
            return self._bridge.groups.zone.get_lights(zone.id)
        return []

    async def set_flash(self, id: str, short: bool = False) -> None:
        """Send Flash command to grouped_light."""
        if short:
            # redirect command to underlying lights
            await asyncio.gather(
                *[
                    self._bridge.lights.set_flash(
                        id=light.id,
                        short=True,
                    )
                    for light in self.get_lights(id)
                ]
            )
            return
        await self.set_state(id, alert=AlertEffectType.BREATHE)

    async def set_state(
        self,
        id: str,
        on: bool | None = None,
        brightness: float | None = None,
        color_xy: tuple[float, float] | None = None,
        color_temp: int | None = None,
        transition_time: int | None = None,
        alert: AlertEffectType | None = None,
    ) -> None:
        """Set supported feature(s) to grouped_light resource."""
        # Sending (color) commands to grouped_light was added in Bridge version 1.50.1950111030
        self._bridge.config.require_version("1.50.1950111030")
        update_obj = GroupedLightPut()
        if on is not None:
            update_obj.on = OnFeature(on=on)
        if brightness is not None:
            update_obj.dimming = DimmingFeaturePut(brightness=brightness)
        if color_xy is not None:
            update_obj.color = ColorFeaturePut(xy=ColorPoint(*color_xy))
        if color_temp is not None:
            update_obj.color_temperature = ColorTemperatureFeaturePut(mirek=color_temp)
        if transition_time is not None:
            update_obj.dynamics = DynamicsFeaturePut(duration=transition_time)
        if alert is not None:
            update_obj.alert = AlertFeaturePut(action=alert)

        await self.update(id, update_obj)

    async def set_dimming_delta(
        self, id: str, brightness_delta: float | None = None
    ) -> None:
        """
        Set brightness_delta value and action via DimmingDeltaFeature.

        The action to be send depends on brightness_delta value:
         None: STOP (this immediately stops any dimming transition)
         > 0: UP,
         < 0: DOWN
        """
        if brightness_delta in (None, 0):
            dimming_delta = DimmingDeltaFeaturePut(action=DeltaAction.STOP)
        else:
            dimming_delta = DimmingDeltaFeaturePut(
                action=DeltaAction.UP if brightness_delta > 0 else DeltaAction.DOWN,
                brightness_delta=abs(brightness_delta),
            )

        update_obj = GroupedLightPut()
        update_obj.dimming_delta = dimming_delta
        await self.update(id, update_obj)


class GroupsController(GroupedControllerBase[Union[Room, Zone, GroupedLight]]):  # noqa: UP007
    """Controller grouping resources of both room and zone."""

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self.grouped_light = GroupedLightController(bridge)
        self.room = RoomController(bridge)
        self.zone = ZoneController(bridge)
        super().__init__(
            bridge,
            [
                self.room,
                self.zone,
                self.grouped_light,
            ],
        )



"""Controller holding and managing HUE resources of type `scene`."""

from typing import TYPE_CHECKING, Any

from aiohue.v2.models.feature import DimmingFeaturePut, RecallAction, RecallFeature
from aiohue.v2.models.resource import ResourceTypes
from aiohue.v2.models.room import Room
from aiohue.v2.models.scene import Scene, ScenePut
from aiohue.v2.models.smart_scene import (
    SmartScene,
    SmartScenePut,
    SmartSceneRecall,
    SmartSceneRecallAction,
)
from aiohue.v2.models.zone import Zone

from .base import BaseResourcesController, GroupedControllerBase

if TYPE_CHECKING:
    from aiohue.v2 import HueBridgeV2

SCENE_TYPES = Scene | SmartScene


class RegularScenesController(BaseResourcesController[type[Scene]]):
    """Controller holding and managing HUE resources of type `scene`."""

    item_type = ResourceTypes.SCENE
    item_cls = Scene
    allow_parser_error = True

    async def recall(
        self,
        id: str,
        dynamic: bool = False,
        duration: int | None = None,
        brightness: float | None = None,
    ) -> None:
        """Turn on / recall scene."""
        action = RecallAction.DYNAMIC_PALETTE if dynamic else RecallAction.ACTIVE
        update_obj = ScenePut(recall=RecallFeature(action=action, duration=duration))
        if brightness is not None:
            update_obj.recall.dimming = DimmingFeaturePut(brightness=brightness)
        await self.update(id, update_obj)

    def get_group(self, id: str) -> Room | Zone:
        """Get group attached to given scene id."""
        scene = self[id]
        return next(x for x in self._bridge.groups if x.id == scene.group.rid)


class SmartScenesController(BaseResourcesController[type[SmartScene]]):
    """Controller holding and managing HUE resources of type `smart_scene`."""

    item_type = ResourceTypes.SMART_SCENE
    item_cls = SmartScene
    allow_parser_error = True

    async def recall(
        self, id: str, action: SmartSceneRecallAction = SmartSceneRecallAction.ACTIVATE
    ) -> None:
        """Turn on / recall scene."""
        update_obj = SmartScenePut(recall=SmartSceneRecall(action=action))
        await self.update(id, update_obj)

    def get_group(self, id: str) -> Room | Zone:
        """Get group attached to given scene id."""
        scene = self[id]
        return next(x for x in self._bridge.groups if x.id == scene.group.rid)


class ScenesController(GroupedControllerBase[SCENE_TYPES]):
    """Controller grouping resources of all scene resources."""

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self.scene = RegularScenesController(bridge)
        self.smart_scene = SmartScenesController(bridge)
        super().__init__(
            bridge,
            [
                self.scene,
                self.smart_scene,
            ],
        )

    async def recall(self, id: str, *args: Any, **kwargs: Any) -> None:
        """Turn on / recall scene."""
        scene = self[id]
        # forward call to correct controller
        # this method is here for convenience (and backwards compatibility)
        if isinstance(scene, SmartScene):
            await self.smart_scene.recall(id, *args, **kwargs)
            return
        await self.scene.recall(id, *args, **kwargs)

    def get_group(self, id: str) -> Room | Zone:
        """Get group attached to given scene id."""
        scene = self[id]
        return next(x for x in self._bridge.groups if x.id == scene.group.rid)



"""Controller holding and managing HUE resources of sensor type."""

import asyncio
from typing import TYPE_CHECKING

from aiohue.util import dataclass_to_dict
from aiohue.v2.models.bell_button import BellButton
from aiohue.v2.models.bridge_home import BridgeHome
from aiohue.v2.models.button import Button, ButtonEvent
from aiohue.v2.models.camera_motion import CameraMotion, CameraMotionPut
from aiohue.v2.models.contact import Contact, ContactPut
from aiohue.v2.models.convenience_area_motion import (
    ConvenienceAreaMotion,
    ConvenienceAreaMotionPut,
)
from aiohue.v2.models.device_power import DevicePower
from aiohue.v2.models.feature import MotionSensingFeatureSensitivityPut
from aiohue.v2.models.geofence_client import GeofenceClient
from aiohue.v2.models.grouped_light_level import GroupedLightLevel, GroupedLightLevelPut
from aiohue.v2.models.grouped_motion import GroupedMotion, GroupedMotionPut
from aiohue.v2.models.light_level import LightLevel, LightLevelPut
from aiohue.v2.models.motion import Motion, MotionPut
from aiohue.v2.models.motion_area_configuration import MotionAreaConfiguration
from aiohue.v2.models.relative_rotary import RelativeRotary
from aiohue.v2.models.resource import ResourceTypes
from aiohue.v2.models.room import Room
from aiohue.v2.models.security_area_motion import (
    SecurityAreaMotion,
    SecurityAreaMotionPut,
)
from aiohue.v2.models.service_group import ServiceGroup
from aiohue.v2.models.tamper import Tamper
from aiohue.v2.models.temperature import Temperature
from aiohue.v2.models.zigbee_connectivity import ZigbeeConnectivity
from aiohue.v2.models.zone import Zone

from .base import BaseResourcesController, GroupedControllerBase
from .events import EventType

if TYPE_CHECKING:
    from aiohue.v2 import HueBridgeV2

SENSOR_TYPES = (
    DevicePower
    | BellButton
    | Button
    | CameraMotion
    | Contact
    | ConvenienceAreaMotion
    | GeofenceClient
    | GroupedLightLevel
    | GroupedMotion
    | LightLevel
    | Motion
    | RelativeRotary
    | SecurityAreaMotion
    | Tamper
    | Temperature
    | ZigbeeConnectivity
)


BTN_WORKAROUND_NEEDED = ("FOHSWITCH",)


class DevicePowerController(BaseResourcesController[type[DevicePower]]):
    """Controller holding and managing HUE resources of type `device_power`."""

    item_type = ResourceTypes.DEVICE_POWER
    item_cls = DevicePower
    allow_parser_error = True


class BellButtonController(BaseResourcesController[type[BellButton]]):
    """Controller holding and managing HUE resources of type `bell_button`."""

    item_type = ResourceTypes.BELL_BUTTON
    item_cls = BellButton
    allow_parser_error = True


class ButtonController(BaseResourcesController[type[Button]]):
    """Controller holding and managing HUE resources of type `button`."""

    item_type = ResourceTypes.BUTTON
    item_cls = Button
    allow_parser_error = True

    _workaround_tasks: dict[str, asyncio.Task] = None

    async def _handle_event(
        self,
        evt_type: EventType,
        evt_data: dict | None,
        is_reconnect: bool = False,  # noqa: ARG002
    ) -> None:
        """Handle incoming event for this resource from the EventStream."""
        await super()._handle_event(evt_type, evt_data)

        if not evt_data:
            return

        # Handle longpress workaround if needed
        if not (
            evt_type == EventType.RESOURCE_UPDATED
            and evt_data.get("button", {}).get("button_report", {}).get("event")
            == ButtonEvent.INITIAL_PRESS.value
        ):
            return

        device = self.get_device(evt_data["id"])
        if device is None or device.product_data.model_id not in BTN_WORKAROUND_NEEDED:
            return

        if self._workaround_tasks is None:
            self._workaround_tasks = {}

        # cancel existing task (if any)
        # should not happen, but just in case
        task = self._workaround_tasks.pop(evt_data["id"], None)
        if task and not task.done():
            task.cancel()

        self._workaround_tasks[evt_data["id"]] = asyncio.create_task(
            self._handle_longpress_workaround(evt_data["id"])
        )

    async def _handle_longpress_workaround(self, id: int):
        """Handle workaround for FOH switches."""
        # Fake `held down` and `long press release` events.
        # This might need to be removed in a future release once/if Signify
        # adds this back in their API.
        self._logger.debug("Long press workaround for FOH switch initiated.")
        btn_resource = dataclass_to_dict(self._items[id])
        await asyncio.sleep(1.5)  # time to initially wait for SHORT_RELEASE
        count = 0
        try:
            while count <= 20:  # = max 10 seconds
                cur_event = self._items[id].button.button_report.event
                if cur_event == ButtonEvent.SHORT_RELEASE:
                    break
                # send REPEAT until short release is received
                btn_resource["button"]["button_report"]["event"] = (
                    ButtonEvent.REPEAT.value
                )
                await self._handle_event(EventType.RESOURCE_UPDATED, btn_resource)
                await asyncio.sleep(0.5)
                count += 1
        finally:
            # Fire LONG_RELEASE event if time between INITIAL_PRESS and SHORT_RELEASE
            # is more than 1.5 seconds or 10 seconds expired.
            # The button will not send SHORT_RELEASE if more than 10 seconds passed.
            # Note that the button will also fire the SHORT_RELEASE event if it's released within
            # those 10 seconds.
            if count > 1:
                btn_resource["button"]["button_report"]["event"] = (
                    ButtonEvent.LONG_RELEASE.value
                )
                await self._handle_event(EventType.RESOURCE_UPDATED, btn_resource)
            self._logger.debug("Long press workaround for FOH switch completed.")


class CameraMotionController(BaseResourcesController[type[CameraMotion]]):
    """Controller holding and managing HUE resources of type `camera_motion`."""

    item_type = ResourceTypes.CAMERA_MOTION
    item_cls = CameraMotion
    allow_parser_error = True

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, MotionPut(enabled=enabled))

    async def set_sensitivity(self, id: str, sensitivity: int) -> None:
        """Enable/Disable sensor."""
        await self.update(id, CameraMotionPut(sensitivity=sensitivity))


class ContactController(BaseResourcesController[type[Contact]]):
    """Controller holding and managing HUE resources of type `contact`."""

    item_type = ResourceTypes.CONTACT
    item_cls = Contact
    allow_parser_error = True

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, ContactPut(enabled=enabled))


class ConvenienceAreaMotionController(
    BaseResourcesController[type[ConvenienceAreaMotion]]
):
    """Controller holding and managing HUE resources of type `convenience_area_motion`."""

    item_type = ResourceTypes.CONVENIENCE_AREA_MOTION
    item_cls = ConvenienceAreaMotion
    allow_parser_error = True

    def get_motion_area_configuration(self, id: str) -> MotionAreaConfiguration:
        """Get the motion area configuration connected to convenience area motion."""
        for area_config in self._bridge.config.motion_area_configuration:
            for service in area_config.services:
                if service.rid == id:
                    return area_config
        raise KeyError(
            f"Motion area configuration for convenience area motion {id} not found"
        )

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, ConvenienceAreaMotionPut(enabled=enabled))

    async def set_sensitivity(self, id: str, sensitivity: int) -> None:
        """Set motion sensor sensitivity."""
        await self.update(
            id,
            ConvenienceAreaMotionPut(
                sensitivity=MotionSensingFeatureSensitivityPut(sensitivity=sensitivity)
            ),
        )


class SecurityAreaMotionController(BaseResourcesController[type[SecurityAreaMotion]]):
    """Controller holding and managing HUE resources of type `security_area_motion`."""

    item_type = ResourceTypes.SECURITY_AREA_MOTION
    item_cls = SecurityAreaMotion
    allow_parser_error = True

    def get_motion_area_configuration(self, id: str) -> MotionAreaConfiguration:
        """Get the motion area configuration connected to security area motion."""
        for area_config in self._bridge.config.motion_area_configuration:
            for service in area_config.services:
                if service.rid == id:
                    return area_config
        raise KeyError(
            f"Motion area configuration for security area motion {id} not found"
        )

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, SecurityAreaMotionPut(enabled=enabled))

    async def set_sensitivity(self, id: str, sensitivity: int) -> None:
        """Set motion sensor sensitivity."""
        await self.update(
            id,
            SecurityAreaMotionPut(
                sensitivity=MotionSensingFeatureSensitivityPut(sensitivity=sensitivity)
            ),
        )


class GeofenceClientController(BaseResourcesController[type[GeofenceClient]]):
    """Controller holding and managing HUE resources of type `geofence_client`."""

    item_type = ResourceTypes.GEOFENCE_CLIENT
    item_cls = GeofenceClient
    allow_parser_error = True


class GroupedLightLevelController(BaseResourcesController[type[GroupedLightLevel]]):
    """Controller holding and managing HUE resources of type `grouped_light_level`."""

    item_type = ResourceTypes.GROUPED_LIGHT_LEVEL
    item_cls = GroupedLightLevel
    allow_parser_error = True

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, GroupedLightLevelPut(enabled=enabled))


class GroupedMotionController(BaseResourcesController[type[GroupedMotion]]):
    """Controller holding and managing HUE resources of type `grouped_motion`."""

    item_type = ResourceTypes.GROUPED_MOTION
    item_cls = GroupedMotion
    allow_parser_error = True

    def get_parent(self, id: str) -> ServiceGroup | Room | Zone | BridgeHome | None:
        """Get the parent (ServiceGroup or Room/Zone) connected to grouped motion (if any)."""
        grouped_motion = self._items[id]
        if grouped_motion.owner.rtype == ResourceTypes.SERVICE_GROUP:
            return self._bridge.config.service_group[grouped_motion.owner.rid]
        if grouped_motion.owner.rtype == ResourceTypes.ROOM:
            return self._bridge.groups.room[grouped_motion.owner.rid]
        if grouped_motion.owner.rtype == ResourceTypes.ZONE:
            return self._bridge.groups.zone[grouped_motion.owner.rid]
        if grouped_motion.owner.rtype == ResourceTypes.BRIDGE_HOME:
            return self._bridge.config.bridge_home
        return None

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, GroupedMotionPut(enabled=enabled))


class LightLevelController(BaseResourcesController[type[LightLevel]]):
    """Controller holding and managing HUE resources of type `light_level`."""

    item_type = ResourceTypes.LIGHT_LEVEL
    item_cls = LightLevel
    allow_parser_error = True

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, LightLevelPut(enabled=enabled))


class MotionController(BaseResourcesController[type[Motion]]):
    """Controller holding and managing HUE resources of type `motion`."""

    item_type = ResourceTypes.MOTION
    item_cls = Motion
    allow_parser_error = True

    async def set_enabled(self, id: str, enabled: bool) -> None:
        """Enable/Disable sensor."""
        await self.update(id, MotionPut(enabled=enabled))


class RelativeRotaryController(BaseResourcesController[type[Button]]):
    """Controller holding and managing HUE resources of type `relative_rotary`."""

    item_type = ResourceTypes.RELATIVE_ROTARY
    item_cls = RelativeRotary
    allow_parser_error = True


class TamperController(BaseResourcesController[type[Tamper]]):
    """Controller holding and managing HUE resources of type `tamper`."""

    item_type = ResourceTypes.TAMPER
    item_cls = Tamper
    allow_parser_error = True


class TemperatureController(BaseResourcesController[type[Temperature]]):
    """Controller holding and managing HUE resources of type `temperature`."""

    item_type = ResourceTypes.TEMPERATURE
    item_cls = Temperature
    allow_parser_error = True


class ZigbeeConnectivityController(BaseResourcesController[type[ZigbeeConnectivity]]):
    """Controller holding and managing HUE resources of type `zigbee_connectivity`."""

    item_type = ResourceTypes.ZIGBEE_CONNECTIVITY
    item_cls = ZigbeeConnectivity
    allow_parser_error = True


class SensorsController(GroupedControllerBase[SENSOR_TYPES]):
    """Controller grouping resources of all sensor resources."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, bridge: "HueBridgeV2") -> None:
        """Initialize instance."""
        self.bell_button = BellButtonController(bridge)
        self.button = ButtonController(bridge)
        self.camera_motion = CameraMotionController(bridge)
        self.contact = ContactController(bridge)
        self.convenience_area_motion = ConvenienceAreaMotionController(bridge)
        self.device_power = DevicePowerController(bridge)
        self.geofence_client = GeofenceClientController(bridge)
        self.grouped_light_level = GroupedLightLevelController(bridge)
        self.grouped_motion = GroupedMotionController(bridge)
        self.light_level = LightLevelController(bridge)
        self.motion = MotionController(bridge)
        self.relative_rotary = RelativeRotaryController(bridge)
        self.security_area_motion = SecurityAreaMotionController(bridge)
        self.tamper = TamperController(bridge)
        self.temperature = TemperatureController(bridge)
        self.zigbee_connectivity = ZigbeeConnectivityController(bridge)
        super().__init__(
            bridge,
            [
                self.bell_button,
                self.button,
                self.camera_motion,
                self.contact,
                self.convenience_area_motion,
                self.device_power,
                self.geofence_client,
                self.grouped_light_level,
                self.grouped_motion,
                self.light_level,
                self.motion,
                self.relative_rotary,
                self.security_area_motion,
                self.tamper,
                self.temperature,
                self.zigbee_connectivity,
            ],
        )



"""
Model(s) for entertainment resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_entertainment
"""

from dataclasses import dataclass

from .entertainment_configuration import Segment
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class SegmentationProperties:
    """
    Represent a SegmentationProperties dict type.

    All properties regarding the segment capabilities of this device:
    the configuratibility, max_segments and all segment tables.
    """

    configurable: bool
    max_segments: int
    segments: list[Segment]


@dataclass
class Entertainment:
    """
    Represent a (full) `Entertainment` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_entertainment_get
    """

    id: str
    owner: ResourceIdentifier
    # renderer: Indicates if a lamp can be used for entertainment streaming as renderer
    renderer: bool
    # proxy: Indicates if a lamp can be used for entertainment streaming as a proxy node
    proxy: bool
    # max_streams: (integer – minimum: 1)
    # Indicates the maximum number of parallel streaming sessions the bridge supports
    max_streams: int | None = 1
    # segments: Holds all parameters concerning the segmentations capabilities of a device
    segments: SegmentationProperties | None = None

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.ENTERTAINMENT



"""
Model(s) for button resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_button
"""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from .resource import ResourceIdentifier, ResourceTypes


class ButtonEvent(Enum):
    """Enum with possible button events."""

    INITIAL_PRESS = "initial_press"
    REPEAT = "repeat"
    SHORT_RELEASE = "short_release"
    LONG_PRESS = "long_press"
    LONG_RELEASE = "long_release"
    DOUBLE_SHORT_RELEASE = "double_short_release"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, value: object):  # noqa: ARG003
        """Set default enum member if an unknown value is provided."""
        return ButtonEvent.UNKNOWN


@dataclass
class ButtonReport:
    """
    Represent ButtonReport as retrieved from api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_button_get
    """

    updated: datetime
    event: ButtonEvent


@dataclass
class ButtonFeature:
    """Represent ButtonFeature object as used by the Hue api."""

    button_report: ButtonReport | None = None
    last_event: ButtonEvent | None = None  # deprecated
    repeat_interval: int | None = None
    event_values: list[ButtonEvent] | None = None

    @property
    def value(self) -> ButtonEvent:
        """Return the actual/current value."""
        # prefer new style attribute (not available on older firmware versions)
        if self.button_report is not None:
            return self.button_report.event
        if self.last_event is not None:
            return self.last_event
        return ButtonEvent.UNKNOWN


@dataclass
class ButtonMetadata:
    """Represent ButtonMetadata object as used by the Button resource."""

    # number of control within switch (value between 0..8).
    # Meaning in combination with type
    # - dots: Number of dots
    # - number: Number printed on device
    # - other: a logical order of controls in switch
    control_id: int


@dataclass
class Button:
    """
    Represent a (full) `Button` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_button_get
    """

    id: str
    owner: ResourceIdentifier
    metadata: ButtonMetadata

    button: ButtonFeature | None = None
    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.BUTTON



"""Common utilities for fixtures."""

import json
import pathlib


def load_fixture(name: str):
    """Load a fixture from disk."""
    path = pathlib.Path(__file__).parent / "fixtures" / name

    content = path.read_text()

    if name.endswith(".json"):
        return json.loads(content)

    return content



"""Test v2 bridge."""

from unittest.mock import patch

from aiohue import HueBridgeV2


async def test_bridge_init(v2_resources):
    """Test v2 bridge."""
    bridge = HueBridgeV2("192.168.1.123", "mock-key")
    assert bridge.host == "192.168.1.123"

    with patch.object(bridge, "request", return_value=v2_resources):
        await bridge.fetch_full_state()

    assert bridge.config is not None
    assert bridge.config.bridge_id == "aabbccddeeffggh"

    assert bridge.devices is not None
    assert len(bridge.devices.get_lights("0b216218-d811-4c95-8c55-bbcda50f9d50")) == 1
    assert len(bridge.devices.get_sensors("342daec9-391b-480b-abdd-87f1aa04ce3b")) == 6

    assert bridge.lights is not None
    assert bridge.scenes is not None
    assert bridge.sensors is not None
    assert bridge.groups is not None

    # test required version check
    assert bridge.config.check_version("1.50.1950111030") is False
    assert bridge.config.check_version("1.48.1948086000") is True
    assert bridge.config.check_version("1.48.1948085000") is True



"""
Model(s) for matter_fabric resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_matter_fabric
"""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from .resource import ResourceTypes


class MatterFabricStatus(Enum):
    """Enum with the possible MatterFabric statuses."""

    PENDING = "pending"
    PAIRED = "paired"
    TIMEDOUT = "timedout"


@dataclass
class MatterFabricData:
    """Human readable context to identify Matter fabric."""

    label: str
    vendor_id: int


@dataclass
class MatterFabric:
    """
    Represent a (full) `MatterFabric` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_matter_fabric_get
    """

    id: str
    status: MatterFabricStatus
    # NOTE: Only a fabric with status paired has fabric_data
    fabric_data: MatterFabricData | None = None
    creation_time: datetime | None = None
    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.MATTER_FABRIC



"""AIOHue example for HUE bridge authentication."""

import argparse
import asyncio
import contextlib

from aiohue import create_app_key

parser = argparse.ArgumentParser(description="AIOHue Authentication Example")
parser.add_argument("host", help="hostname of Hue bridge")
args = parser.parse_args()


async def main():
    """Run code example."""
    host = args.host
    print()
    print("Creating app_key for bridge: ", args.host)
    print()
    # the link button must be pressed before sending the authentication request
    input("Press the link button on the bridge and press enter to continue...")

    # request api_key from bridge
    try:
        api_key = await create_app_key(host, "authentication_example")
        print()
        print("Authentication succeeded, api key: ", api_key)
        print("NOTE: store the app_key for next connections, it does not expire.")
        print()
    except Exception as exc:  # pylint: disable=broad-except
        print("ERROR: ", str(exc))
        print()


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""
Model(s) for Zone resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zone
"""

from dataclasses import dataclass

from .resource import ResourceTypes
from .room import Room, RoomPost, RoomPut


@dataclass
class Zone(Room):
    """
    Represent a (full) `Zone` object as retrieved from the Hue api.

    Zones group services and each service can be part of multiple zones.
    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zone_get
    """

    type: ResourceTypes = ResourceTypes.ZONE


@dataclass
class ZonePut(RoomPut):
    """
    Properties to send when updating/setting a `Zone` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zone__id__put
    """


@dataclass
class ZonePost(RoomPost):
    """
    Properties to send when creating a `Zone` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_zone_post
    """



"""Test parser functions that converts the incoming json from API into dataclass models."""

import datetime
from dataclasses import dataclass

import pytest

from aiohue.util import dataclass_from_dict


@dataclass
class BasicModelChild:
    """Basic test model."""

    a: int
    b: str
    c: str
    d: int | None


@dataclass
class BasicModel:
    """Basic test model."""

    a: int
    b: float
    c: str
    d: int | None
    e: BasicModelChild
    f: datetime.datetime
    g: str = "default"


def test_dataclass_from_dict():
    """Test dataclass from dict parsing."""
    raw = {
        "a": 1,
        "b": 1.0,
        "c": "hello",
        "d": 1,
        "e": {"a": 2, "b": "test", "c": "test", "d": None},
        "f": "2022-12-09T06:58:00Z",
    }
    res = dataclass_from_dict(BasicModel, raw)
    # test the basic values
    assert isinstance(res, BasicModel)
    assert res.a == 1
    assert res.b == 1.0
    assert res.d == 1
    # test recursive parsing
    assert isinstance(res.e, BasicModelChild)
    # test default value
    assert res.g == "default"
    # test int gets converted to float
    raw["b"] = 2
    res = dataclass_from_dict(BasicModel, raw)
    assert res.b == 2.0
    # test datetime string
    assert isinstance(res.f, datetime.datetime)
    assert res.f.month == 12
    assert res.f.day == 9
    # test string doesn't match int
    with pytest.raises(TypeError):
        raw2 = {**raw}
        raw2["a"] = "blah"
        dataclass_from_dict(BasicModel, raw2)
    # test missing key result in keyerror
    with pytest.raises(KeyError):
        raw2 = {**raw}
        del raw2["a"]
        dataclass_from_dict(BasicModel, raw2)
    # test extra keys silently ignored in non-strict mode
    raw2 = {**raw}
    raw2["extrakey"] = "something"
    dataclass_from_dict(BasicModel, raw2, strict=False)
    # test extra keys not silently ignored in strict mode
    with pytest.raises(KeyError):
        dataclass_from_dict(BasicModel, raw2, strict=True)



"""
Model(s) for Scene resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene
"""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from .feature import (
    ColorFeatureBase,
    ColorTemperatureFeatureBase,
    DimmingFeatureBase,
    DynamicsFeaturePut,
    GradientFeatureBase,
    OnFeature,
    PaletteFeature,
    RecallFeature,
    SceneEffectsFeature,
)
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class ActionAction:
    """Represent (scene) `ActionAction` model."""

    on: OnFeature | None = None
    dimming: DimmingFeatureBase | None = None
    color: ColorFeatureBase | None = None
    color_temperature: ColorTemperatureFeatureBase | None = None
    gradient: GradientFeatureBase | None = None
    effects: SceneEffectsFeature | None = None
    dynamics: DynamicsFeaturePut | None = None


@dataclass
class Action:
    """Represent (scene) `Action` model."""

    target: ResourceIdentifier
    action: ActionAction


@dataclass
class SceneMetadata:
    """Represent SceneMetadata object as used by the Hue api."""

    name: str
    image: ResourceIdentifier | None = None


class SceneActiveStatus(Enum):
    """Enum with possible active statuses for a Scene.

    Hue API docs (resource: scene / status.active):
    - inactive
    - static
    - dynamic_palette
    """

    INACTIVE = "inactive"
    STATIC = "static"
    DYNAMIC_PALETTE = "dynamic_palette"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, value: object):  # noqa: ARG003
        """Return default member if unknown value encountered."""
        return SceneActiveStatus.UNKNOWN


@dataclass
class SceneStatus:
    """Represent Scene Status object.

    Consists of information about the current status and last time the scene was recalled.
    Hue API fields:
        - active: one of inactive | static | dynamic_palette
        - last_recall: ISO 8601 timestamp (optional if never recalled)
    """

    active: SceneActiveStatus
    last_recall: datetime | None = None


@dataclass
class SceneMetadataPut:
    """Represent SceneMetadata model when sent/updated to the API with PUT request."""

    name: str


@dataclass
class Scene:
    """
    Represent (full) `Scene` Model when retrieved from the API.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene__id__get
    """

    id: str
    metadata: SceneMetadata
    # group: required(object)
    # Group associated with this Scene. All services in the group are part of this scene.
    # If the group is changed the scene is updated (e.g. light added/removed)
    group: ResourceIdentifier
    # actions: required(array of Action)
    # List of actions to be executed synchronously on recall
    actions: list[Action]
    # speed: required(number – minimum: 0 – maximum: 1)
    speed: float
    # auto_dynamic: whether to automatically start the scene dynamically on active recall
    auto_dynamic: bool | None = None
    # status: the current active status and last recall of the scene
    status: SceneStatus | None = None

    # optional params
    id_v1: str | None = None
    palette: PaletteFeature | None = None

    type: ResourceTypes = ResourceTypes.SCENE


@dataclass
class ScenePut:
    """
    Properties to send when updating/setting a `Scene` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene__id__put
    """

    metadata: SceneMetadataPut | None = None
    actions: list[Action] | None = None
    palette: PaletteFeature | None = None
    recall: RecallFeature | None = None
    palette: PaletteFeature | None = None
    # speed: (number – minimum: 0 – maximum: 1)
    # Speed of dynamic palette for this scene
    speed: float | None = None
    auto_dynamic: bool | None = None


@dataclass
class ScenePost:
    """
    Properties to send when creating a `Scene` object on the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_scene_post
    """

    metadata: SceneMetadata
    group: ResourceIdentifier
    actions: list[Action]
    palette: PaletteFeature | None = None
    speed: float | None = None
    auto_dynamic: bool | None = None
    type: ResourceTypes = ResourceTypes.SCENE



"""
Model(s) for tamper resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_tamper
"""

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum

from .resource import ResourceIdentifier, ResourceTypes


class TamperState(Enum):
    """State of a Tamper sensor."""

    TAMPERED = "tampered"
    NOT_TAMPERED = "not_tampered"


class TamperSource(Enum):
    """Source of a Tamper alert."""

    BATTERY_DOOR = "battery_door"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, value: object):  # noqa: ARG003
        """Set default enum member if an unknown value is provided."""
        return TamperSource.UNKNOWN


@dataclass
class TamperReport:
    """
    Represent TamperReport as retrieved from api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_tamper_get
    """

    changed: datetime
    source: TamperSource
    state: TamperState


@dataclass
class Tamper:
    """
    Represent a (full) `Tamper` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_tamper_get
    """

    id: str
    owner: ResourceIdentifier
    tamper_reports: list[TamperReport] = field(default_factory=list)

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.CONTACT



"""
Model(s) for grouped_motion resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_motion
"""

from dataclasses import dataclass

from .feature import (
    MotionSensingFeature,
)
from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class GroupedMotion:
    """
    Represent a (full) `GroupedMotion` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_motion_get
    """

    id: str
    owner: ResourceIdentifier
    # enabled: required(boolean)
    # true when sensor is activated, false when deactivated
    enabled: bool
    motion: MotionSensingFeature

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.GROUPED_MOTION


@dataclass
class GroupedMotionPut:
    """
    GroupedMotion resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_grouped_motion__id__put
    """

    enabled: bool | None = None
    type: ResourceTypes = ResourceTypes.GROUPED_MOTION



"""
Model(s) for light_level resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_level
"""

from dataclasses import dataclass
from datetime import datetime

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class LightLevelReport:
    """
    Represent LightLevelReport as retrieved from api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_level_get
    """

    changed: datetime
    light_level: int


@dataclass
class LightLevelFeature:
    """Represent LightLevel Feature used by Lightlevel resources."""

    light_level_report: LightLevelReport | None
    light_level: int | None = None  # deprecated
    light_level_valid: bool | None = None  # deprecated

    @property
    def value(self) -> int | None:
        """Return the actual/current value."""
        # prefer new style attribute (not available on older firmware versions)
        if self.light_level_report is not None:
            return self.light_level_report.light_level
        return self.light_level


@dataclass
class LightLevel:
    """
    Represent a (full) `LightLevel` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_level_get
    """

    id: str
    owner: ResourceIdentifier
    enabled: bool
    light: LightLevelFeature

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.LIGHT_LEVEL


@dataclass
class LightLevelPut:
    """
    LightLevel resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_light_level__id__put
    """

    enabled: bool | None = None



"""
Model(s) for bell_button resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_bell_button
"""

from dataclasses import dataclass


from aiohue.v2.models.button import ButtonFeature, ButtonMetadata

from .resource import ResourceIdentifier, ResourceTypes


@dataclass
class BellButton:
    """
    Represent a (full) `BellButton` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_bell_button_get
    """

    id: str
    owner: ResourceIdentifier
    metadata: ButtonMetadata

    button: ButtonFeature | None = None
    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.BELL_BUTTON



"""Example script for using AIOHue connecting to a V2 Hue bridge."""

import argparse
import asyncio
import contextlib
import logging

from aiohue import HueBridgeV2

parser = argparse.ArgumentParser(description="AIOHue Example")
parser.add_argument("host", help="hostname of Hue bridge")
parser.add_argument("appkey", help="appkey for Hue bridge")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main():
    """Run Main execution."""
    if args.debug:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(asctime)-15s %(levelname)-5s %(name)s -- %(message)s",
        )

    async with HueBridgeV2(args.host, args.appkey) as bridge:
        print("Connected to bridge: ", bridge.bridge_id)
        print(bridge.config.bridge_device)

        print()
        print("found devices:")
        for item in bridge.devices:
            print(item.metadata.name)

        # turn on a light
        light = next(x for x in bridge.lights.items if x.supports_color)
        print("Turning on light", light.id)
        await bridge.lights.turn_on(light.id)
        await asyncio.sleep(1)
        print("Set brightness 100 to light", light.id)
        await bridge.lights.set_brightness(light.id, 100, 2000)
        await asyncio.sleep(2)
        print("Set color to light", light.id)
        await bridge.lights.set_color(light.id, 0.141, 0.123, 2000)
        await asyncio.sleep(1)
        print("Turning off light", light.id)
        await bridge.lights.turn_off(light.id, 2000)

        print()
        print("Subscribing to events...")

        def print_event(event_type, item):
            print()
            print("received event", event_type.value, item)
            print()

        bridge.subscribe(print_event)

        await asyncio.sleep(3600)


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""
Model(s) for contact resource on HUE bridge.

https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_contact
"""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from .resource import ResourceIdentifier, ResourceTypes


class ContactState(Enum):
    """State of a Contact sensor."""

    CONTACT = "contact"
    NO_CONTACT = "no_contact"


@dataclass
class ContactReport:
    """
    Represent ContactReport as retrieved from api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_contact_get
    """

    changed: datetime
    state: ContactState


@dataclass
class Contact:
    """
    Represent a (full) `Contact` resource when retrieved from the api.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_contact_get
    """

    id: str
    owner: ResourceIdentifier
    # enabled: required(boolean)
    # true when sensor is activated, false when deactivated
    enabled: bool
    contact_report: ContactReport | None = None

    id_v1: str | None = None
    type: ResourceTypes = ResourceTypes.CONTACT


@dataclass
class ContactPut:
    """
    Contact resource properties that can be set/updated with a PUT request.

    https://developers.meethue.com/develop/hue-api-v2/api-reference/#resource_contact__id__put
    """

    enabled: bool | None = None



"""Test v1 bridge."""

from aiohue import HueBridgeV1


async def test_bridge_init():
    """Test v1 bridge."""
    bridge = HueBridgeV1("192.168.1.123", "mock-key")
    assert bridge.host == "192.168.1.123"



"""Test parsing of Scene.status field."""

import datetime
from unittest.mock import patch

from aiohue import HueBridgeV2
from aiohue.v2.models.scene import SceneActiveStatus


async def test_scene_status_parsing(v2_resources):
    """Ensure scene status (active + last_recall) is parsed correctly."""
    bridge = HueBridgeV2("127.0.0.1", "fake")

    with patch.object(bridge, "request", return_value=v2_resources):
        await bridge.fetch_full_state()

    # Collect scenes
    scenes = list(bridge.scenes.scene)  # regular scenes controller
    assert len(scenes) >= 2  # fixture contains at least two scenes

    dynamic_scene = scenes[0]
    static_scene = scenes[1]

    # dynamic scene assertions
    assert dynamic_scene.status is not None
    assert dynamic_scene.status.active == SceneActiveStatus.DYNAMIC_PALETTE
    assert isinstance(dynamic_scene.status.last_recall, datetime.datetime)

    # static scene assertions
    assert static_scene.status is not None
    assert static_scene.status.active == SceneActiveStatus.STATIC
    assert isinstance(static_scene.status.last_recall, datetime.datetime)



"""Example script for using AIOHue connecting to a V1 Hue bridge."""

import argparse
import asyncio
import contextlib
import logging

from aiohue import HueBridgeV1
from aiohue.v1.lights import Light

parser = argparse.ArgumentParser(description="AIOHue Example")
parser.add_argument("host", help="hostname of Hue bridge")
parser.add_argument("appkey", help="appkey for Hue bridge")
parser.add_argument("--debug", help="enable debug logging", action="store_true")
args = parser.parse_args()


async def main():
    """Run Main execution."""
    if args.debug:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(asctime)-15s %(levelname)-5s %(name)s -- %(message)s",
        )

    async with HueBridgeV1(args.host, args.appkey) as bridge:
        print("Connected to bridge: ", bridge.bridge_id)

        print()
        print("found lights:")
        for item in bridge.lights:
            print(item)

        # turn on a light
        light: Light = bridge.lights.items[0]
        print("Turning on light", light.name)
        await light.set_state(on=True)
        await asyncio.sleep(1)

        print("Set brightness 100 to light", light.name)
        await light.set_state(bri=100, transitiontime=1000)
        await asyncio.sleep(2)

        print("Set color to light", light.name)
        await light.set_state(xy=(0.141, 0.123))
        await asyncio.sleep(1)
        print("Turning off light", light.name)
        await light.set_state(on=False)

        # V1 bridge does not support events
        # you must request updates yourself manually...
        print("requesting update of all lights")
        await bridge.lights.update()


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""AIOHue example for HUE bridge discovery."""

import asyncio
import contextlib

from aiohue.discovery import discover_nupnp


async def main():
    """Run code example."""
    discovered_bridges = await discover_nupnp()
    print()
    print("Discovered bridges:")
    for item in discovered_bridges:
        print()
        print(item)
        print()


with contextlib.suppress(KeyboardInterrupt):
    asyncio.run(main())



"""Test base controller functions."""

from dataclasses import dataclass
from unittest.mock import Mock
from uuid import uuid4

from aiohue import HueBridgeV2
from aiohue.v2 import EventType
from aiohue.v2.controllers.base import BaseResourcesController
from aiohue.v2.models.resource import ResourceTypes


@dataclass
class MockData:
    """Mock data resource type."""

    id: str
    type: ResourceTypes = ResourceTypes.UNKNOWN

    id_v1: str | None = None


class MockController(BaseResourcesController[type[MockData]]):
    """Controller for mock data resource."""

    item_type = ResourceTypes.UNKNOWN
    item_cls = MockData
    allow_parser_error = False


async def test_handle_event():
    """Test handling of events."""
    bridge = HueBridgeV2("127.0.0.1", "fake")
    controller = MockController(bridge)
    callback = Mock(return_value=None)
    controller.subscribe(callback)

    resource_id = str(uuid4())
    other_id = str(uuid4())

    evt_data = {
        "id": resource_id,
        "id_v1": "mock/1",
    }

    # Create a new resource
    # pylint: disable=protected-access
    await controller._handle_event(EventType.RESOURCE_ADDED, evt_data)

    cur_data = MockData(
        id=resource_id,
        type=ResourceTypes.UNKNOWN,
        id_v1="mock/1",
    )
    callback.assert_called_once_with(EventType.RESOURCE_ADDED, cur_data)
    callback.reset_mock()

    evt_data = {
        "id": resource_id,
        "id_v1": "mock/2",
    }

    # Update of a single property of an existing resource
    # pylint: disable=protected-access
    await controller._handle_event(EventType.RESOURCE_UPDATED, evt_data)

    cur_data = MockData(
        id=resource_id,
        type=ResourceTypes.UNKNOWN,
        id_v1="mock/2",
    )
    callback.assert_called_once_with(EventType.RESOURCE_UPDATED, cur_data)
    callback.reset_mock()

    evt_data = {
        "id": other_id,
        "id_v1": "mock/1",
    }

    # Update of a single property of a non-existing resource
    # pylint: disable=protected-access
    await controller._handle_event(EventType.RESOURCE_UPDATED, evt_data)

    callback.assert_not_called()
    callback.reset_mock()

    evt_data = {
        "id": resource_id,
    }

    # Remove of existing resource
    # pylint: disable=protected-access
    await controller._handle_event(EventType.RESOURCE_DELETED, evt_data)

    cur_data = MockData(
        id=resource_id,
        type=ResourceTypes.UNKNOWN,
        id_v1="mock/2",
    )
    callback.assert_called_once_with(EventType.RESOURCE_DELETED, cur_data)
    callback.reset_mock()

    evt_data = {
        "id": resource_id,
        "id_v1": "mock/2",
    }

    # Update of an already removed resource
    # pylint: disable=protected-access
    await controller._handle_event(EventType.RESOURCE_UPDATED, evt_data)

    callback.assert_not_called()
